from py4j import java_gateway
from pyboof import gateway


def is_java_class(java_class, string_path):
    return gateway.jvm.pyboof.PyBoofEntryPoint.isClass(java_class,string_path)


def is_java_list(object):
    object.getClass()
    return True


class JavaWrapper:
    def __init__(self, java_object=None):
        self.java_obj = java_object
        self.java_fields = [x for x in gateway.jvm.pyboof.PyBoofEntryPoint.getPublicFields(self.java_obj.getClass())]

    def __getattr__(self, item):
        if "java_fields" in self.__dict__ and item in self.__dict__["java_fields"]:
            return java_gateway.get_field(self.java_obj, item)
        else:
            return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        if "java_fields" in self.__dict__ and key in self.__dict__["java_fields"]:
            java_gateway.set_field(self.java_obj, key, value)
        else:
            self.__dict__[key] = value

    def __dir__(self):
        return sorted(set( self.__dict__.keys() + self.java_fields))

    def set_java_object(self, obj):
        self.java_obj = obj

    def get_java_object(self):
        return self.java_obj

    def __str__(self):
        return "Wrapped Java: "+self.java_obj.toString()


class JavaConfig(JavaWrapper):
    """
    Provides a nice python wrapper around Java classes.  Public variables are automatically turned into Python
    attributes
    """
    # TODO variables which are java classes are a little messed up
    def __init__(self, java_class_path ):
        if isinstance(java_class_path, basestring):
            self.java_class_path = java_class_path

            words = java_class_path.replace('$',".").split(".")

            a = gateway.jvm.__getattr__(words[0])
            for i in range(1,len(words)):
                a = a.__getattr__(words[i])

            self.java_obj = a.__call__()
        else:
            self.java_obj = java_class_path
        JavaWrapper.__init__(self, self.java_obj)

    def __getattr__(self, item):
        if "java_fields" in self.__dict__ and item in self.__dict__["java_fields"]:
            a = java_gateway.get_field(self.java_obj,item)
            if gateway.jvm.pyboof.PyBoofEntryPoint.isConfigClass(a):
                return JavaConfig(a)
            else:
                return a
        else:
            return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        if "java_fields" in self.__dict__ and key in self.__dict__["java_fields"]:
            if isinstance(value, JavaConfig):
                java_gateway.set_field(self.java_obj, key, value.java_obj)
            else:
                java_gateway.set_field(self.java_obj, key, value)
        else:
            self.__dict__[key] = value


class JavaList(JavaWrapper):
    def __init__(self, java_list, java_type):
        JavaWrapper.__init__(self,java_list)
        self.java_type = java_type

    def size(self):
        return self.java_obj.size()

    def save_to_disk(self, file_name ):
        gateway.jvm.pyboof.FileIO.saveList(self.java_obj,self.java_type,file_name)


def JavaList_to_fastqueue(java_list, java_class_type , queue_declare=False):
    return gateway.jvm.pyboof.PyBoofEntryPoint.listToFastQueue(java_list, java_class_type, queue_declare)




import atexit
import mmap
import os
import signal
import subprocess
import time

from py4j.java_gateway import JavaGateway
from py4j.protocol import Py4JError
from py4j.protocol import Py4JNetworkError

gateway = JavaGateway()

mmap_size = 0
mmap_file = None

build_date = None

java_pid = None

# Read the date everything was build
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),"build_date.txt"), 'r') as f:
    build_date = f.readline()

if build_date is None:
    print "Can't find build_data.txt at "+os.path.dirname(os.path.realpath(__file__))
    exit(1)

def check_jvm( set_date ):
    global gateway
    try:
        gateway.jvm.pyboof.PyBoofEntryPoint.nothing()
        if set_date:
            gateway.jvm.pyboof.PyBoofEntryPoint.setBuildDate(build_date)
        else:
            java_build_date = gateway.jvm.pyboof.PyBoofEntryPoint.getBuildDate()
            if build_date != java_build_date:
                print "Python and Java build dates do not match.  Killing Java process."
                print "  build dates = {:s} {:s}".format(build_date,java_build_date)
                gateway.close()
                time.sleep(1)
                return False

    except Py4JNetworkError:
        return False
    except Py4JError as e:
        print e
        print "Py4J appears to have attached itself to a process that doesn't have the expected jars.  " \
              "Try killing py4j processes"
        exit(1)
    return True


def shutdown_jvm():
    global java_pid, gateway
    if java_pid is None:
        pass
    else:
        # shutdown the gateway so that it doesn't spew out a billion error messages when it can't connect
        # to the JVM
        gateway.shutdown()
        gateway = None
        os.kill(java_pid, signal.SIGTERM)
        java_pid = None

# Catch control-c and kill the java process "gracefully" first.
def signal_handler(signal, frame):
    shutdown_jvm()
    sys.exit(0)
signal.signal(signal.SIGINT, signal_handler)

# kill java on a regular exit too
atexit.register(shutdown_jvm)

if not check_jvm(False):
    print "Launching Java process"
    jar_path = os.path.realpath(__file__)
    jar_path = os.path.join(os.path.dirname(jar_path),"PyBoof-all.jar")
    proc = subprocess.Popen(["java","-jar",jar_path])
    java_pid = proc.pid
    time.sleep(0.1)
    # closed loop initialization.  If it fails for 5 seconds give up
    start_time = time.time()
    success = False
    while time.time() - start_time < 5.0:
        if check_jvm(True):
            success = True
            break

    if not success:
        print "Failed to successfully launch the JVM after 5 seconds.  Aborting"
        pass

class MmapType:
    """
    Type byte for different data structures
    """
    IMAGE_U8 = 0
    IMAGE_F32 = 1
    LIST_POINT2D_F64 = 2
    LIST_TUPLE_F64 = 3


def init_memmap( size_MB=2):
    """
    Call to enable use of memory mapped files for quick communication between Python and Java.  This
    faster communication method requires specialized code so is only used when large amounts of memory
    is being transferred.

    :param size_MB: Size of the memory mapped file in megabytes
    :type size_MB: int
    """
    global mmap_size, mmap_file
    mmap_name = "mmap_python_java.mmap"
    mmap_size = size_MB*1024*1024
    mmap_name_path = os.path.join(os.getcwd(), mmap_name)
    gateway.jvm.pyboof.PyBoofEntryPoint.initializeMmap(mmap_name_path, size_MB)
    # Open file in read,write,binary mode
    mmap_fid = open(mmap_name, "r+b")
    mmap_file = mmap.mmap(mmap_fid.fileno(), length=0, flags=mmap.MAP_SHARED,
                          prot=mmap.PROT_READ | mmap.PROT_WRITE)

from calib import *
from common import *
from geo import *
from image import *
from ip import *
from recognition import *
from feature import *
from stereo import *
import swing



#!/usr/bin/env python

import unittest

from pyboof import gateway
import pyboof as pb

pb.init_memmap()


class TestMemMapFunctions(unittest.TestCase):
    def test_convert_list_tuple64(self):
        original = [[1, 2.34, 6.7], [-23.4, 934.123, 56.1]]

        # python to java
        java_list = gateway.jvm.java.util.ArrayList()
        pb.mmap_list_python_to_TupleF64(original, java_list)

        # java to python
        found = []
        pb.mmap_list_TupleF64_to_python(java_list, found)

        self.assertEqual(len(original), len(found))

        for i in xrange(len(original)):
            a = original[i]
            b = found[i]
            self.assertEqual(len(a),len(b))
            for j in xrange(len(a)):
                self.assertEqual(a[j],b[j])

    def test_convert_list_Point2DF64(self):
        original = [(1, 2.34), (-23.4, 934.123)]

        # python to java
        java_list = gateway.jvm.java.util.ArrayList()
        pb.mmap_list_python_to_Point2DF64(original, java_list)

        # java to python
        found = []
        pb.mmap_list_Point2DF64_to_python(java_list, found)

        self.assertEqual(len(original), len(found))

        for i in xrange(len(original)):
            a = original[i]
            b = found[i]
            self.assertEquals(a[0], b[0])
            self.assertEquals(a[1], b[1])


if __name__ == '__main__':
    unittest.main()



import struct
import sys
import pyboof
from pyboof import JavaConfig
from pyboof import JavaWrapper
from pyboof import dtype_to_Class_SingleBand
from pyboof import gateway
from common import JavaList
from common import JavaList_to_fastqueue
from common import is_java_class


def p2b_list_descF64(pylist):
    """
    Converts a python list of feature descriptors stored in 64bit floats into a BoofCV compatible format
    :param pylist: Python list of feature descriptors
    :type pylist: list[list[float]]
    :return: List of descriptors in BoofCV format
    """
    java_list = gateway.jvm.java.util.ArrayList()

    if pyboof.mmap_file:
        mmap_list_python_to_TupleF64(pylist, java_list)
    else:
        raise Exception("Yeah this needs to be implemented.  Turn mmap on if possible")
    return java_list

def b2p_list_descF64(boof_list):
    """
    Converts a BoofCV list of feature descriptors stored in 64bit floats into a Python compatible format
    :param boof_list: Descriptor list in BoofCV format
    :return: List of descriptors in Python format
    :rtype: list[list[float]]
    """
    pylist = []

    if pyboof.mmap_file:
        mmap_list_TupleF64_to_python(boof_list, pylist)
    else:
        raise Exception("Yeah this needs to be implemented.  Turn mmap on if possible")
    return pylist

def p2b_list_point2DF64( pylist ):
    """
    Converts a python list of feature descriptors stored in 64bit floats into a BoofCV compatible format
    :param pylist: Python list of 2d points
    :type pylist: list[(float,float)]
    :return: List of 2d points in BoofCV format
    """
    java_list = gateway.jvm.java.util.ArrayList()

    if pyboof.mmap_file:
        mmap_list_python_to_Point2DF64(pylist,java_list)
    else:
        raise Exception("Yeah this needs to be implemented.  Turn mmap on if possible")
    return java_list


def b2p_list_point2DF64( boof_list ):
    """
    Converts a BoofCV list of 2d points into a Python compatible format
    :param boof_list: Descriptor list in BoofCV format
    :return: List of 2d points in Python format
    :type pylist: list[(float,float)]
    """
    pylist = []

    if pyboof.mmap_file:
        mmap_list_Point2DF64_to_python(boof_list,pylist)
    else:
        raise Exception("Yeah this needs to be implemented.  Turn mmap on if possible")
    return pylist


class ConfigSurfFast(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.abst.feature.describe.ConfigSurfDescribe$Speed")


class ConfigSurfStability(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.abst.feature.describe.ConfigSurfDescribe$Stability")


class ConfigFastHessian(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.abst.feature.detect.interest.ConfigFastHessian")


class ConfigAverageIntegral(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.abst.feature.orientation.ConfigAverageIntegral")


class ConfigDenseSampling(JavaConfig):
    def __init__(self, scale,periodX ,periodY):
        JavaConfig.__init__(self,"boofcv.factory.feature.dense.ConfigDenseSample")
        self.scale = scale
        self.periodX = periodX
        self.periodY = periodY


class AssocScoreType:
    """
    Enum different types of association scoring techniques
    """
    DEFAULT = 0,
    SAD = 1
    EUCLIDEAN = 2
    EUCLIDEAN_SQ = 3
    NCC = 4


class ConfigAssociation:
    def __init__(self,score_type=AssocScoreType.DEFAULT, max_error=sys.float_info.max,backwards_validation=True):
        self.score_type = score_type
        self.max_error = max_error
        self.backwards_validation = backwards_validation


class AssociateDescription(JavaWrapper):
    def __init__(self, java_object ):
        JavaWrapper.__init__(self, java_object)

    def set_source(self, feature_list):
        """

        :param feature_list: List of feature descriptions
        :type feature_list: [[float]] | JavaList
        """
        # automatically convert from python to boof type
        if type(feature_list) is list:
            feature_list = p2b_list_descF64(feature_list)
        elif type(feature_list) is JavaList:
            feature_list = feature_list.java_obj
        else:
            raise Exception("unexpected list type "+feature_list.__class__.__name__)

        java_type = gateway.jvm.boofcv.struct.feature.TupleDesc_F64(0).getClass()

        fast_queue = JavaList_to_fastqueue(feature_list, java_type, queue_declare=False)
        self.java_obj.setSource(fast_queue)

    def set_destination(self, feature_list):
        """

        :param feature_list: List of feature descriptions
        :type feature_list: [[float]] | JavaList
        """
        # automatically convert from python to boof type
        if type(feature_list) is list:
            feature_list = p2b_list_descF64(feature_list)
        elif type(feature_list) is JavaList:
            feature_list = feature_list.java_obj
        else:
            raise Exception("unexpected list type "+feature_list.__class__.__name__)

        java_type = gateway.jvm.boofcv.struct.feature.TupleDesc_F64(0).getClass()

        fast_queue = JavaList_to_fastqueue(feature_list, java_type, queue_declare=False)
        self.java_obj.setDestination(fast_queue)

    def associate(self):
        """
        Associates the two sets of features together.  Returns a list of association indexes
        :return: List of (src index, dst index, match score)
        :rtype: [(int,int,float)]
        """
        output = []

        self.java_obj.associate()
        matches = self.java_obj.getMatches()
        for i in xrange(matches.getSize()):
            association = matches.get(i)
            output.append( (association.src,association.dst,association.score) )

        return output

    def get_java_matches(self):
        return self.java_obj.getMatches()


def read_list_tuple_desc_f64(f, list_length):
    output = []
    for i in xrange(list_length):
        desc_length = struct.unpack('>i', f.read(4))[0]
        desc = [0.0]*desc_length
        for j in xrange(desc_length):
            desc[j] = struct.unpack('>d', f.read(8))[0]
        output.append(desc)
    return output


def read_list(file_name):
    output = []
    with open(file_name, 'r') as f:
        data_type = f.readline()[0:-1]
        if data_type != "list":
            raise RuntimeError("Was expecting list in front of file not "+data_type)
        class_type = f.readline()
        list_length = struct.unpack('>i', f.read(4))[0]

        if "TupleDesc_F64" in class_type:
            output = read_list_tuple_desc_f64(f,list_length)
        else:
            raise RuntimeError("Unknown list data type "+class_type)

    if list_length != len(output):
        raise RuntimeError("Unexpected list size. "+str(list_length)+" "+str(len(output)))
    return output


def java_list_to_python(java_list):
    N = java_list.size()
    output = []
    if is_java_class(java_list.java_type,"boofcv.struct.feature.TupleDesc_F64"):
        for i in xrange(N):
            d = java_list.java_obj.get(i)
            value = d.getValue() # some hackery to get around py4j short comings
            output.append([x for x in value])
    elif is_java_class(java_list.java_type,"georegression.struct.point.Point2D_F64"):
        for i in xrange(N):
            p = java_list.java_obj.get(i)
            output.append((p.x,p.y))
    else:
        raise RuntimeError("Unknown java list type")

    return output

class DetectDescribePointFeatures(JavaWrapper):
    def __init__(self,java_object):
        self.set_java_object(java_object)

    def detect(self, image ):
        """
        Detects features inside the image and returns a list of feature locations and descriptions
        :param image: Image in BoofCV format
        :return: List of feature pixel locations and their descriptions. (2d pixel coordinates, list of descriptions)
        :rtype: (list[(float,float)],list[list[float]])
        """
        self.java_obj.detect(image)

        # extract a list of locations and descriptions.  Don't copy since it will immediately be convert
        # into a python format
        java_locations = gateway.jvm.pyboof.PyBoofEntryPoint.extractPoints(self.java_obj,False)
        java_descriptions = gateway.jvm.pyboof.PyBoofEntryPoint.extractFeatures(self.java_obj,False)

        # Convert into a Python format and return the two lists
        locations = b2p_list_point2DF64(java_locations)
        descriptions = b2p_list_descF64(java_descriptions)

        return locations, descriptions

    def get_scales(self):
        N = self.java_obj.getNumberOfFeatures()
        output = [0]*N
        for i in xrange(N):
            output[i] = self.java_obj.getScale(i)
        return output

    def get_orientations(self ):
        N = self.java_obj.getNumberOfFeatures()
        output = [0]*N
        for i in xrange(N):
            output[i] = self.java_obj.getOrientation(i)
        return output

    def has_scale(self):
        return self.java_obj.hasScale()

    def has_orientation(self):
        return self.java_obj.hasOrientation()

    def get_descriptor_type(self):
        return self.java_obj.getDescriptionType()


class DenseDescribePointFeatures(JavaWrapper):
    def __init__(self,java_object):
        self.set_java_object(java_object)

    def detect(self, image ):
        pass


class FactoryDetectDescribe:
    def __init__(self, dtype ):
        self.boof_image_type =  dtype_to_Class_SingleBand(dtype)

    def createSurf( self, config_detect=None , config_desc=None , config_ori=None ):
        """
        Creates a SURF detector and describer.

        :param config_detect:
        :type config_detect: ConfigFastHessian | None
        :param config_desc:
        :type config_desc: ConfigSurfFast | ConfigSurfStability | None
        :param config_ori:
        :type config_ori: ConfigAverageIntegral
        :return:
        :rtype: DetectDescribePointFeatures
        """
        if config_desc is None:
            config_desc = ConfigSurfStability()

        java_config_detect = None
        java_config_desc = config_desc.java_obj
        java_config_ori = None

        if config_detect is not None:
            java_config_detect = config_detect.java_obj
        if config_ori is not None:
            java_config_ori = config_ori.java_obj

        if config_desc.__class__.__name__ == "ConfigSurfFast":
            java_object = gateway.jvm.boofcv.factory.feature.detdesc.FactoryDetectDescribe.surfFast(
                java_config_detect,java_config_desc,java_config_ori, self.boof_image_type)
        elif config_desc.__class__.__name__ == "ConfigSurfStability":
            java_object = gateway.jvm.boofcv.factory.feature.detdesc.FactoryDetectDescribe.surfStable(
                java_config_detect,java_config_desc,java_config_ori, self.boof_image_type)
        else:
            raise RuntimeError("Unknown description type")

        return DetectDescribePointFeatures(java_object)

    def createSift(self):
        pass

    def createBrief(self):
        pass

    def createBriefSO(self):
        pass

    def createNcc(self):
        pass


class FactoryDenseDescribe:
    def __init__(self, dtype ):
        self.boof_image_type =  dtype_to_Class_SingleBand(dtype)

    def createSurf( self, config_sampling , config_desc=None ):
        if config_desc is None:
            config_desc = ConfigSurfFast()

        java_config_sampling = config_sampling.java_obj
        java_config_desc = config_desc.java_obj


class FactoryAssociate:
    def __init__(self):
        self.score = None

    def set_score(self, score_type , descriptor_type ):
        if score_type == AssocScoreType.DEFAULT:
            self.score = gateway.jvm.boofcv.factory.feature.associate.\
                FactoryAssociation.defaultScore(descriptor_type)
        elif score_type == AssocScoreType.EUCLIDEAN:
            self.score = gateway.jvm.boofcv.factory.feature.associate.\
                FactoryAssociation.scoreEuclidean(descriptor_type,False)
        elif score_type == AssocScoreType.EUCLIDEAN_SQ:
            self.score = gateway.jvm.boofcv.factory.feature.associate.\
                FactoryAssociation.scoreEuclidean(descriptor_type,True)
        elif score_type == AssocScoreType.NCC:
            self.score = gateway.jvm.boofcv.factory.feature.associate.\
                FactoryAssociation.scoreNcc()
        elif score_type == AssocScoreType.SAD:
            self.score = gateway.jvm.boofcv.factory.feature.associate.\
                FactoryAssociation.scoreSad(descriptor_type)

    def greedy(self, max_error=sys.float_info.max,backwards_validation=True):
        java_obj = gateway.jvm.boofcv.factory.feature.associate.\
                FactoryAssociation.greedy(self.score,max_error,backwards_validation)
        return AssociateDescription(java_obj)

    def kdtree(self):
        pass

    def kdRandomForest(self):
        pass


def mmap_list_python_to_TupleF64(pylist, java_list):
    """
    Converts a python list of float arrays into a list of TupleDesk64F in java using memmap file
    :param pylist: (Input) Python list of float arrays.  All arrays need to have the same length
    :type pylist: list[list[float]]
    :param java_list: (Output) Java list to store TupleDesc64F
    """
    num_elements = len(pylist)
    if num_elements == 0:
        dof = 0
    else:
        dof = len(pylist[0])
    mm = pyboof.mmap_file

    # max number of list elements it can write at once
    max_elements = (pyboof.mmap_size-100)/(dof*8)

    curr = 0
    while curr < num_elements:
        # Write as much of the list as it can to the mmap file
        num_write = min(max_elements,num_elements-curr)
        mm.seek(0)
        mm.write(struct.pack('>HII', pyboof.MmapType.LIST_TUPLE_F64, num_elements, dof))
        for i in range(curr, curr+num_write):
            mm.write(struct.pack('>%sd' % dof, *pylist[i]))

        # Now tell the java end to read what it just wrote
        gateway.jvm.pyboof.PyBoofEntryPoint.mmap.read_List_TupleF64(java_list)

        # move on to the next block
        curr = curr + num_write

def mmap_list_TupleF64_to_python(java_list, pylist):
    """
    Converts a java list of TupleDesc64F into a python list of float arrays using memmap file
    :param java_list: Input: java list
    :param pylist: output: python list
    :type pylist: list[list[float]]
    """
    num_elements = java_list.size()
    mm = pyboof.mmap_file

    num_read = 0
    while num_read < num_elements:
        gateway.jvm.pyboof.PyBoofEntryPoint.mmap.write_List_TupleF64(java_list, num_read)
        mm.seek(0)
        data_type, num_found, dof = struct.unpack(">HII", mm.read(2+4+4))
        if data_type != pyboof.MmapType.LIST_TUPLE_F64:
            raise Exception("Unexpected data type in mmap file. {%d}" % data_type)
        if num_found > num_elements-num_read:
            raise Exception("Too many elements returned. "+str(num_found))
        for i in xrange(num_found):
            desc = struct.unpack(">%sd" % dof, mm.read(8*dof))
            pylist.append(desc)
        num_read += num_found

def mmap_list_python_to_Point2DF64(pylist, java_list):
    """
    Converts a python list of 2d float tuples into a list of Point2D_64F in java using memmap file

    :param pylist: (Input) Python list of 2D float tuples.
    :type pylist: list[(float,float)]
    :param java_list: (Output) Java list to store Point2D_64F
    """
    num_elements = len(pylist)
    mm = pyboof.mmap_file

    # max number of list elements it can write at once
    max_elements = (pyboof.mmap_size-100)/(2*8)

    curr = 0
    while curr < num_elements:
        # Write as much of the list as it can to the mmap file
        num_write = min(max_elements,num_elements-curr)
        mm.seek(0)
        mm.write(struct.pack('>HI', pyboof.MmapType.LIST_POINT2D_F64, num_elements))
        for i in range(curr, curr+num_write):
            mm.write(struct.pack('>2d', *pylist[i]))

        # Now tell the java end to read what it just wrote
        gateway.jvm.pyboof.PyBoofEntryPoint.mmap.read_List_Point2DF64(java_list)

        # move on to the next block
        curr = curr + num_write

def mmap_list_Point2DF64_to_python( java_list , pylist ):
    """
    Converts a java list of Point2D_F64 into a python list of float 2D tuples using memmap file
    :param java_list: Input: java list
    :param pylist: output: python list
    :type pylist: list[(float,float)]
    """
    num_elements = java_list.size()
    mm = pyboof.mmap_file

    num_read = 0
    while num_read < num_elements:
        gateway.jvm.pyboof.PyBoofEntryPoint.mmap.write_List_Point2DF64(java_list, num_read)
        mm.seek(0)
        data_type, num_found = struct.unpack(">HI", mm.read(2+4))
        if data_type != pyboof.MmapType.LIST_POINT2D_F64:
            raise Exception("Unexpected data type in mmap file. %d" % data_type)
        if num_found > num_elements-num_read:
            raise Exception("Too many elements returned. "+str(num_found))
        for i in xrange(num_found):
            desc = struct.unpack(">2d", mm.read(8*2))
            pylist.append(desc)
        num_read += num_found




#!/usr/bin/env python

import unittest

from pyboof import gateway
import pyboof as pb
import numpy as np

pb.init_memmap()


class TestBImage(unittest.TestCase):

    dtypes = (np.uint8, np.float32)

    def test_array_read_gray(self):
        for dtype in self.dtypes:
            j_img = pb.create_single_band(100, 120, dtype=dtype)
            b_image = pb.BImage(j_img)

            self.assertEqual(j_img.get(3,2), b_image[2,3])

    def test_array_read_planar(self):
        for dtype in self.dtypes:
            j_img = pb.create_planar(20, 30, 3, dtype=dtype)
            b_image = pb.BImage(j_img)

            self.assertEqual(j_img.getBand(1).get(3,2), b_image[2,3,1])

    def test_array_read_interleaved(self):
        for dtype in self.dtypes:
            j_img = pb.create_interleaved(20, 30, 3, dtype=dtype)
            b_image = pb.BImage(j_img)

            self.assertEqual(j_img.getBand(3,2,1), b_image[2,3,1])

    def test_array_write_gray(self):
        for dtype in self.dtypes:
            j_img = pb.create_single_band(100, 120, dtype=dtype)
            b_image = pb.BImage(j_img)

            b_image[2,3] = 5

            self.assertEqual(5, j_img.get(3,2))

    def test_array_write_planar(self):
        for dtype in self.dtypes:
            j_img = pb.create_planar(20, 30, 3, dtype=dtype)
            b_image = pb.BImage(j_img)

            b_image[2, 3, 0] = 5

            self.assertEqual(5, j_img.getBand(0).get(3, 2))

    def test_array_write_interleaved(self):
        for dtype in self.dtypes:
            j_img = pb.create_interleaved(20, 30, 3, dtype=dtype)
            b_image = pb.BImage(j_img)

            b_image[2, 3, 0] = 5

            self.assertEqual(5, j_img.getBand(3, 2, 0))

    def test_property_reading(self):
        j_img = pb.create_single_band(100, 120, dtype=np.uint8)
        b_image = pb.BImage(j_img)
        self.assertEqual(100, b_image.width)
        self.assertEqual(120, b_image.height)

    def test_shape_gray(self):
        j_img = pb.create_single_band(100, 120, dtype=np.uint8)
        found = pb.BImage(j_img).shape
        self.assertEqual(2, len(found))
        self.assertEqual(120, found[0])
        self.assertEqual(100, found[1])

    def test_shape_planar(self):
        j_img = pb.create_planar(100, 120, 3, dtype=np.uint8)
        found = pb.BImage(j_img).shape
        self.assertEqual(3, len(found))
        self.assertEqual(120, found[0])
        self.assertEqual(100, found[1])
        self.assertEqual(3, found[2])

    def test_shape_interleaved(self):
        j_img = pb.create_interleaved(100, 120, 3, dtype=np.uint8)
        found = pb.BImage(j_img).shape
        self.assertEqual(3, len(found))
        self.assertEqual(120, found[0])
        self.assertEqual(100, found[1])
        self.assertEqual(3, found[2])


class TestMemMapFunctions(unittest.TestCase):

    def test_mmap_numpy_to_boof_U8(self):
        np_gray = np.random.randint(0, 256, size=(100, 120), dtype=np.uint8)
        pb_gray = pb.mmap_numpy_to_boof_U8(np_gray)

        self.assertEqual( np_gray.shape[0], pb_gray.getHeight())
        self.assertEqual( np_gray.shape[1], pb_gray.getWidth())

        self.assertEqual(np_gray[0, 0], pb_gray.get(0, 0))
        self.assertEqual(np_gray[20, 10], pb_gray.get(10, 20))

    def test_mmap_numpy_to_boof_F32(self):
        np_gray = np.random.random((100, 120)).astype(np.float32)
        pb_gray = pb.mmap_numpy_to_boof_F32(np_gray)

        self.assertEqual(np_gray.shape[0], pb_gray.getHeight())
        self.assertEqual(np_gray.shape[1], pb_gray.getWidth())

        self.assertAlmostEqual(np_gray[0, 0], pb_gray.get(0, 0))
        self.assertAlmostEqual(np_gray[20, 10], pb_gray.get(10, 20))

    def test_mmap_numpy_to_boof_IU8(self):
        np_img = np.random.randint(0,256, size=(100, 120, 3), dtype=np.uint8)
        pb_img = pb.mmap_numpy_to_boof_IU8(np_img)

        self.assertEqual(np_img.shape[0], pb_img.getHeight())
        self.assertEqual(np_img.shape[1], pb_img.getWidth())
        self.assertEqual(np_img.shape[2], pb_img.getNumBands())

        self.assertEqual(np_img[0, 0, 0], pb_img.getBand(0, 0, 0))
        self.assertEqual(np_img[20, 10, 0], pb_img.getBand(10, 20, 0))
        self.assertEqual(np_img[20, 10, 1], pb_img.getBand(10, 20, 1))
        self.assertEqual(np_img[20, 10, 2], pb_img.getBand(10, 20, 2))

    def test_mmap_boof_to_numpy_U8(self):
        pb_img = pb.create_single_band(100, 120, dtype=np.uint8)
        pb.fill_uniform(pb_img, 0, 200)
        np_img = pb.mmap_boof_to_numpy_U8(pb_img)

        self.assertEquals(np_img.dtype, np.uint8)
        self.assertEqual(np_img.shape[0], pb_img.getHeight())
        self.assertEqual(np_img.shape[1], pb_img.getWidth())

        self.assertAlmostEqual(np_img[0, 0], pb_img.get(0, 0))
        self.assertAlmostEqual(np_img[20, 10], pb_img.get(10, 20))

    def test_mmap_boof_to_numpy_F32(self):
        pb_img = pb.create_single_band(100, 120, dtype=np.float32)
        pb.fill_uniform(pb_img, -2, 2)
        np_img = pb.mmap_boof_to_numpy_F32(pb_img)

        self.assertEquals(np_img.dtype, np.float32)
        self.assertEqual(np_img.shape[0], pb_img.getHeight())
        self.assertEqual(np_img.shape[1], pb_img.getWidth())

        self.assertAlmostEqual(np_img[0, 0], pb_img.get(0, 0))
        self.assertAlmostEqual(np_img[20, 10], pb_img.get(10, 20))

    def test_mmap_boof_PU8_to_numpy_IU8(self):
        pb_img = pb.create_planar(100, 120, 3,dtype=np.uint8)
        pb.fill_uniform(pb_img, -2, 2)
        np_img = pb.mmap_boof_PU8_to_numpy_IU8(pb_img)

        self.assertEquals(np_img.dtype, np.uint8)
        self.assertEqual(np_img.shape[0], pb_img.getHeight())
        self.assertEqual(np_img.shape[1], pb_img.getWidth())
        self.assertEqual(np_img.shape[2], pb_img.getNumBands())

        self.assertEqual(np_img[0, 0, 0]  , pb_img.getBand(0).get(0, 0))
        self.assertEqual(np_img[20, 10, 0], pb_img.getBand(0).get(10, 20))
        self.assertEqual(np_img[20, 10, 1], pb_img.getBand(1).get(10, 20))
        self.assertEqual(np_img[20, 10, 2], pb_img.getBand(2).get(10, 20))

if __name__ == '__main__':
    unittest.main()



import math
import py4j.java_gateway as jg
from pyboof import gateway
import pyboof

def show( boof_image , title="Image"):
    gateway.jvm.boofcv.gui.image.ShowImages.showWindow(boof_image,title)

def show_grid( images , columns=-1, title="Image Grid"):
    if type(images) is not tuple or not list:
        images = (images)

    array = gateway.new_array(gateway.jvm.java.awt.image.BufferedImage,len(images))
    for idx,image in enumerate(images):
        if jg.is_instance_of(gateway, image, gateway.jvm.java.awt.image.BufferedImage ):
            array[idx] = image
        else:
            array[idx] = gateway.jvm.boofcv.io.image.ConvertBufferedImage.convertTo(image,None,True)

    # If no grid is specified try to make it square
    if columns <= 0:
        columns = int(math.sqrt(len(images)))
    gateway.jvm.boofcv.gui.image.ShowImages.showGrid(columns, title, array)


def show_list(image_name_pairs, title="Image List"):
    if type(image_name_pairs) is not tuple or not list:
        image_name_pairs = (image_name_pairs)

    names = []
    buffered = []
    for pair in image_name_pairs:
        if jg.is_instance_of(gateway, pair[0], gateway.jvm.java.awt.image.BufferedImage ):
            buffered.append(pair[0])
        else:
            buffered.append( gateway.jvm.boofcv.io.image.ConvertBufferedImage.convertTo(pair[0],None,True) )
        names.append(pair[1])

    panel = gateway.jvm.boofcv.gui.ListDisplayPanel()
    for i in range(len(names)):
        panel.addImage(buffered[i],names[i])
    gateway.jvm.boofcv.gui.image.ShowImages.showWindow(panel, title)


def colorize_gradient( derivX , derivY ):
    return gateway.jvm.boofcv.gui.image.VisualizeImageData.colorizeGradient(derivX,derivY,-1)


def render_binary( binary , invert=False):
    return gateway.jvm.boofcv.gui.binary.VisualizeBinaryData.renderBinary(binary,invert,None)


def visualize_matches(image_left , image_right , points_src , points_dst , match_indexes,
                      title = "Associated Features"):

    if type(points_src) is list:
        points_src = pyboof.p2b_list_point2DF64(points_src)

    if type(points_dst) is list:
        points_dst = pyboof.p2b_list_point2DF64(points_dst)

    if type(match_indexes) is list:
        raise Exception("Haven't bothered to implement python to java conversion for match_indexes yet")

    # If possible, convert images into a BufferedImage data type
    if jg.is_instance_of(gateway, image_left, gateway.jvm.boofcv.struct.image.ImageBase):
        image_left = gateway.jvm.boofcv.io.image.ConvertBufferedImage.convertTo(image_left, None, True)

    if jg.is_instance_of(gateway, image_right, gateway.jvm.boofcv.struct.image.ImageBase):
        image_right = gateway.jvm.boofcv.io.image.ConvertBufferedImage.convertTo(image_right, None, True)

    panel = gateway.jvm.boofcv.gui.feature.AssociationPanel(20)
    panel.setImages( image_left, image_right)
    panel.setAssociation(points_src, points_dst, match_indexes)
    gateway.jvm.boofcv.gui.image.ShowImages.showWindow(panel, title)





import numpy as np
import os
import struct

import py4j.java_gateway as jg
import pyboof
from pyboof import gateway

import common
from common import JavaWrapper


class Family:
    """
    Enum for the family which the image data type belongs in.

    Equivalent to boofcv.struct.image.ImageType.Family
    """
    SINGLE_BAND = 0
    PLANAR = 1
    INTERLEAVED = 2


class BImage(JavaWrapper):
    """
    Wrapper around a BoofCV image.  Provide a slow but pythonic way to interact with the images.
    """
    def __init__(self, java_image):
        JavaWrapper.__init__(self,java_image)
        self.family = self.java_obj.getImageType().getFamily().ordinal()
        self.dtype = get_dtype(java_image)

    def get_image_type(self):
        """
        Returns the type of image it is
        :return: ImageType
        :rtype: ImageType
        """
        return ImageType(self.java_obj.getImageType())

    def ndarray(self):
        """
        Creates a new ndarray which will have the same values as this image
        :return: equivalent ndarray
        :rtype: numpy.ndarray
        """
        return boof_to_ndarray(self.java_obj)

    def create_same_shape(self):
        """
        Return an image of the same type and shape
        :return: image
        :rtype: BImage
        """
        java_image = self.java_obj.createSameShape()
        return BImage(java_image)

    def __getitem__(self, key):
        if isinstance(key, (list, tuple)):
            if len(key) == 2:
                if self.family is Family.SINGLE_BAND:
                    return self.java_obj.get(key[1], key[0])
                else:
                    raise RuntimeError("2D array like access is only valid for gray scale images")
            elif len(key) == 3:
                if self.family is Family.PLANAR:
                    return self.java_obj.getBand(key[2]).get(key[1], key[0])
                elif self.family is Family.INTERLEAVED:
                    return self.java_obj.getBand(key[2], key[1], key[0])
                else:
                    raise RuntimeError("Must be a multi band image")
            else:
                raise RuntimeError("Unexpected argument length")
        else:
            raise RuntimeError("Unexpected argument type")

    def __setitem__(self, key, value):
        if isinstance(key, (list, tuple)):
            if len(key) == 2:
                if self.family == Family.SINGLE_BAND:
                    if self.dtype == np.uint8:
                        return self.java_obj.set(key[1], key[0], int(value))
                    elif self.dtype == np.float32:
                        return self.java_obj.set(key[1], key[0], float(value))
                else:
                    raise RuntimeError("2D array like access is only valid for gray scale images")
            elif len(key) == 3:
                if self.family == Family.PLANAR:
                    if self.dtype == np.uint8:
                        return self.java_obj.getBand(key[2]).set(key[1], key[0], int(value))
                    elif self.dtype == np.float32:
                        return self.java_obj.getBand(key[2]).set(key[1], key[0], float(value))
                elif self.family == Family.INTERLEAVED:
                    if self.dtype == np.uint8:
                        return self.java_obj.setBand(key[1], key[0], key[2], int(value))
                    elif self.dtype == np.float32:
                        return self.java_obj.setBand(key[1], key[0], key[2], float(value))
                else:
                    raise RuntimeError("Must be a multi band image")
            else:
                raise RuntimeError("Unexpected argument length")
        else:
            raise RuntimeError("Unexpected argument type")

    def __getattr__(self, item):
        if item is 'shape':
            if self.family == Family.SINGLE_BAND:
                return self.java_obj.getHeight(), self.java_obj.getWidth()
            elif self.family == Family.PLANAR or self.family == Family.INTERLEAVED:
                return self.java_obj.getHeight(), self.java_obj.getWidth(), self.java_obj.getNumBands()
        return JavaWrapper.__getattr__(self, item)


class ImageType(JavaWrapper):
    """
    Description on the image data format.

    Equivalent to boofcv.struct.image.ImageType
    """
    def __init__(self, jImageType):
        JavaWrapper.__init__(self, jImageType)

    def create_boof_image(self, width, height):
        return self.java_obj.createImage(width, height)

    def get_family(self):
        return self.java_obj.getFamily().ordinal()

    def get_num_bands(self):
        return self.java_obj.getNumBands()

    def get_dtype(self):
        return JImageDataType_to_dtype(self.java_obj)


def create_ImageType(family, dtype, num_bands=1):
    """
    Returns an ImageType from the specified parameters

    :param family: Image data structure
    :type family: Family
    :param dtype: primitive data type of pixel
    :type dtype: int
    :param num_bands: Number of bands in the image.
    :type num_bands: int
    :return: Returns the ImageType
    :rtype: ImageType
    """
    if family == Family.SINGLE_BAND and num_bands != 1:
        raise Exception("SingleBand images must have only one band")

    j_image_class = dtype_to_Class_SingleBand(dtype)
    if family == Family.SINGLE_BAND:
        j_image_type = gateway.jvm.boofcv.struct.image.ImageType.single(j_image_class)
    elif family == Family.PLANAR:
        j_image_type = gateway.jvm.boofcv.struct.image.ImageType.ms(num_bands, j_image_class)
    elif family == Family.INTERLEAVED:
        j_image_type = gateway.jvm.boofcv.struct.image.ImageType.interleaved(num_bands, j_image_class)
    else:
        raise Exception("Unknown family = "+str(family))

    return ImageType(j_image_type)


def load_single_band( path , dtype ):
    """
    Loads a singe band BoofCV image
    :param path: File path to image
    :param dtype: The data type of the image
    :return:
    """
    file_path = os.path.abspath(path)

    boof_type = dtype_to_Class_SingleBand(dtype)
    found = gateway.jvm.boofcv.io.image.UtilImageIO.loadImage(file_path,boof_type)
    if found is None:
        print file_path
        raise Exception("Can't find image or image format can't be read")
    return found


def load_planar( path , dtype ):
    file_path = os.path.abspath(path)

    buffered_image =  gateway.jvm.boofcv.io.image.UtilImageIO.loadImage(file_path)
    num_bands = buffered_image.getRaster().getNumBands()
    PLANAR = create_planar(buffered_image.getWidth(), buffered_image.getHeight(),num_bands, dtype)
    gateway.jvm.boofcv.io.image.ConvertBufferedImage.convertFrom(buffered_image,PLANAR, True)

    return PLANAR


def convert_boof_image( input , output ):
    """
    Converts between two different boofcv images.  Essentially performs a typecast and doesn't change
    pixel values.  Except if the input has more than one band and the output is gray scale.  It then
    averages the bands.

    :param input: BoofCV image
    :param output:  BoofCV image
    :return:
    """
    gateway.jvm.boofcv.core.image.GConvertImage.convert(input,output)


def ndarray_to_boof( npimg , boof_img=None):
    """
    Converts an image in ndarray format into a BoofCV image

    :param npimg: numpy image
    :param boof_img: Optional storage for BoofCV image.  None to declare a new image
    :return: Converted BoofCV image
    """
    if npimg is None:
        raise Exception("Input image is None")

    if pyboof.mmap_file:
        if len(npimg.shape) == 2:
            if npimg.dtype == np.uint8:
                return mmap_numpy_to_boof_U8(npimg, boof_img)
            elif npimg.dtype == np.float32:
                return mmap_numpy_to_boof_F32(npimg, boof_img)
            else:
                raise Exception("Image type not supported yet")
        else:
            if npimg.dtype == np.uint8:
                return mmap_numpy_to_boof_IU8(npimg)
            elif npimg.dtype == np.float32:
                raise Exception("Need to add support for float images")
            else:
                raise Exception("Image type not supported yet")
    else:
        if len(npimg.shape) == 2:
            if npimg.dtype == np.uint8:
                if boof_img is None:
                    b = gateway.jvm.boofcv.struct.image.GrayU8()
                else:
                    b = boof_img
                b.setData( bytearray(npimg.data))
            elif npimg.dtype == np.float32:
                if boof_img is None:
                    b = gateway.jvm.boofcv.struct.image.GrayF32()
                else:
                    b = boof_img
                b.setData(npimg.data)
            else:
                raise Exception("Image type not supported yet")
        else:
            if boof_img is not None:
                raise RuntimeError("multiband images doesn't yet support predeclared storage")

            # TODO change this to interleaved images since that's the closest equivalent
            num_bands = npimg.shape[2]

            bands = []
            for i in range(num_bands):
                bands.append(npimg[:, :, i])

            class_type = dtype_to_Class_SingleBand(npimg.dtype)

            b = gateway.jvm.boofcv.struct.image.Planar(class_type,num_bands)

            for i in range(num_bands):
                if npimg.dtype == np.uint8:
                    band = gateway.jvm.boofcv.struct.image.GrayU8()
                    band.setData( bytearray(bands[i]) )
                elif npimg.dtype == np.float32:
                    band = gateway.jvm.boofcv.struct.image.GrayF32()
                    band.setData( bands[i] )
                band.setWidth(npimg.shape[1])
                band.setHeight(npimg.shape[0])
                band.setStride(npimg.shape[1])
                b.setBand(i,band)

    b.setWidth(npimg.shape[1])
    b.setHeight(npimg.shape[0])
    b.setStride(npimg.shape[1])

    return b


def boof_to_ndarray( boof ):
    width = boof.getWidth()
    height = boof.getHeight()
    if jg.is_instance_of(gateway, boof, gateway.jvm.boofcv.struct.image.ImageGray):
        nptype = JImageDataType_to_dtype(boof.getImageType().getDataType())
        boof_data = boof.getData()

        if pyboof.mmap_file:
            if nptype == np.uint8:
                return mmap_boof_to_numpy_U8(boof)
            elif nptype == np.float32:
                return mmap_boof_to_numpy_F32(boof)
            else:
                raise RuntimeError("Unsupported image type")
        else:
            # create copy the very slow way
            N = len(boof_data)
            print "Before painful copy {}".format(N)
            data = [0]*N
            for i in xrange(N):
                data[i] = boof_data[i]
            print "After painful copy"
            return np.ndarray(shape=(height,width), dtype=nptype, buffer=np.array(data))
    elif jg.is_instance_of(gateway, boof, gateway.jvm.boofcv.struct.image.Planar):
        nptype = JImageDataType_to_dtype(boof.getImageType().getDataType())
        if pyboof.mmap_file:
            if nptype == np.uint8:
                return mmap_boof_PU8_to_numpy_IU8(boof)
            else:
                raise RuntimeError("Unsupported image type.  Only U8 currently supported")
        else:
            raise RuntimeError("Must have mmap turned on for this image type")
    else:
        raise Exception("Boof image type not yet supported")


def gradient_dtype( dtype ):
    """
    Returns the appropriate image dtype of the provided image dtype
    :param dtype: Type of input image data
    :return: Appropriate image type to store gradient data
    """
    if dtype == np.uint8:
        return np.int16
    elif dtype == np.int8:
        return np.int16
    elif dtype == np.uint16:
        return np.int32
    elif dtype == np.int16:
        return np.int32
    elif dtype == np.float32:
        return np.float32
    elif dtype == np.float64:
        return np.float64
    else:
        raise Exception("Unknown type: "+str(dtype))


def create_single_band(width, height, dtype):
    """
    Creates a single band BoofCV image.

    :param width: Image width
    :param height: Image height
    :param dtype: data type
    :return: New instance of a BoofCV single band image
    """
    if dtype == np.uint8:
        return gateway.jvm.boofcv.struct.image.GrayU8(width, height)
    elif dtype == np.int8:
        return gateway.jvm.boofcv.struct.image.GrayS8(width, height)
    elif dtype == np.uint16:
        return gateway.jvm.boofcv.struct.image.GrayU16(width, height)
    elif dtype == np.int16:
        return gateway.jvm.boofcv.struct.image.GrayS16(width, height)
    elif dtype == np.int32:
        return gateway.jvm.boofcv.struct.image.GrayS32(width, height)
    elif dtype == np.int64:
        return gateway.jvm.boofcv.struct.image.GrayS64(width, height)
    elif dtype == np.float32:
        return gateway.jvm.boofcv.struct.image.GrayF32(width, height)
    elif dtype == np.float64:
        return gateway.jvm.boofcv.struct.image.GrayF64(width, height)
    else:
        raise Exception("Unsupported type")


def create_planar( width , height , num_bands , dtype):
    """
    Creates a Planar BoofCV image.

    :param width: Image width
    :param height: Image height
    :param num_bands: Number of bands in the image
    :param dtype: data type
    :return: New instance of a BoofCV planar multi-bandimage
    """

    jImageClass = dtype_to_Class_SingleBand(dtype)

    return gateway.jvm.boofcv.struct.image.Planar(jImageClass,width,height,num_bands)


def create_interleaved(width, height, num_bands, dtype):
    """
    Creates a interleaved BoofCV image.

    :param width: Image width
    :param height: Image height
    :param num_bands: Number of bands/channels in the image
    :param dtype: data type
    :return: New instance of a BoofCV single band image
    """
    if dtype == np.uint8:
        return gateway.jvm.boofcv.struct.image.InterleavedU8(width, height, num_bands)
    elif dtype == np.int8:
        return gateway.jvm.boofcv.struct.image.InterleavedS8(width, height, num_bands)
    elif dtype == np.uint16:
        return gateway.jvm.boofcv.struct.image.InterleavedU16(width, height, num_bands)
    elif dtype == np.int16:
        return gateway.jvm.boofcv.struct.image.InterleavedS16(width, height, num_bands)
    elif dtype == np.int32:
        return gateway.jvm.boofcv.struct.image.InterleavedS32(width, height, num_bands)
    elif dtype == np.int64:
        return gateway.jvm.boofcv.struct.image.InterleavedS64(width, height, num_bands)
    elif dtype == np.float32:
        return gateway.jvm.boofcv.struct.image.InterleavedF32(width, height, num_bands)
    elif dtype == np.float64:
        return gateway.jvm.boofcv.struct.image.InterleavedF64(width, height, num_bands)
    else:
        raise Exception("Unsupported type")

def get_dtype( boof_image ):
    """
    Given a BoofCV image return the dtype which matches the storage format of its pixels
    :param boof_image: A BoofCV Java image
    :return: The NumPy dtype
    """
    return ImageDataType_to_dtype(boof_image.getImageType().getDataType())


def JImageDataType_to_dtype( ImageDataType ):
    """
    Given an instance of the BoofCV ImageType object return the dtype which represents its
    internal data
    :param ImageType:
    :return:
    """
    if ImageDataType.isInteger():
        if ImageDataType.getNumBits() == 8:
            if ImageDataType.isSigned():
                return np.int8
            else:
                return np.uint8
        elif ImageDataType.getNumBits() == 16:
            if ImageDataType.isSigned():
                return np.int16
            else:
                return np.uint16
        elif ImageDataType.getNumBits() == 32:
            if ImageDataType.isSigned():
                return np.int32
            else:
                raise Exception("Unsigned 32bit data isn't supported")
        elif ImageDataType.getNumBits() == 64:
            if ImageDataType.isSigned():
                return np.int64
            else:
                raise Exception("Unsigned 64bit data isn't supported")
        else:
            raise Exception("Number of bits not supported for ints. "+str(ImageDataType.getNumBits()))
    else:
       if ImageDataType.getNumBits() == 32:
           return np.float32
       elif ImageDataType.getNumBits() == 64:
           return np.float64
       else:
           raise Exception("Number of bits not supported for floats. "+str(ImageDataType.getNumBits()))


def dtype_to_ImageDataType( dtype ):
    if dtype == np.uint8:
        return gateway.jvm.boofcv.struct.image.ImageDataType.U8
    elif dtype == np.int8:
        return gateway.jvm.boofcv.struct.image.ImageDataType.S8
    elif dtype == np.uint16:
        return gateway.jvm.boofcv.struct.image.ImageDataType.U16
    elif dtype == np.int16:
        return gateway.jvm.boofcv.struct.image.ImageDataType.S16
    elif dtype == np.int32:
        return gateway.jvm.boofcv.struct.image.ImageDataType.S32
    elif dtype == np.int64:
        return gateway.jvm.boofcv.struct.image.ImageDataType.S64
    elif dtype == np.float32:
        return gateway.jvm.boofcv.struct.image.ImageDataType.F32
    elif dtype == np.float64:
        return gateway.jvm.boofcv.struct.image.ImageDataType.F64
    else:
        raise Exception("No BoofCV equivalent")


def family_to_Java_Family( family ):
    if family == Family.SINGLE_BAND:
        return gateway.jvm.boofcv.struct.image.ImageType.Family.SINGLE_BAND
    elif family == Family.PLANAR:
        return gateway.jvm.boofcv.struct.image.ImageType.Family.PLANAR
    elif family == Family.INTERLEAVED:
        return gateway.jvm.boofcv.struct.image.ImageType.Family.INTERLEAVED
    else:
        raise Exception("Unknown family. "+str(family))


def dtype_to_Class_SingleBand( dtype ):
    if dtype == np.uint8:
        class_path = "boofcv.struct.image.GrayU8"
    elif dtype == np.int8:
        class_path = "boofcv.struct.image.GrayS8"
    elif dtype == np.uint16:
        class_path = "boofcv.struct.image.GrayU16"
    elif dtype == np.int16:
        class_path = "boofcv.struct.image.GrayS16"
    elif dtype == np.int32:
        class_path = "boofcv.struct.image.GrayS32"
    elif dtype == np.int64:
        class_path = "boofcv.struct.image.GrayS64"
    elif dtype == np.float32:
        class_path = "boofcv.struct.image.GrayF32"
    elif dtype == np.float64:
        class_path = "boofcv.struct.image.GrayF64"
    else:
        raise Exception("No BoofCV equivalent. "+str(dtype))

    return gateway.jvm.java.lang.Class.forName(class_path)


def ImageDataType_to_dtype( jdatatype ):
    if jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.U8:
        return np.uint8
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.S8:
        return np.int8
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.U16:
        return np.uint16
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.S16:
        return np.int16
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.S32:
        return np.int32
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.S64:
        return np.int64
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.F32:
        return np.float32
    elif jdatatype == gateway.jvm.boofcv.struct.image.ImageDataType.F64:
        return np.float64
    else:
        raise Exception("Unknown ImageDataType. "+str(jdatatype))


def ClassSingleBand_to_dtype( jclass ):
    jdatatype = gateway.jvm.boofcv.struct.image.ImageDataType.classToType(jclass)
    return ImageDataType_to_dtype(jdatatype)


def dtype_to_ImageType( dtype ):
    java_class = dtype_to_Class_SingleBand(dtype)
    return gateway.jvm.boofcv.struct.image.ImageType.single(java_class)


def fill_uniform(image, min_value, max_value):
    java_random = gateway.jvm.java.util.Random()
    gateway.jvm.boofcv.alg.misc.GImageMiscOps.fillUniform(image, java_random, float(min_value), float(max_value))


# ================================================================
#        Functions for converting images using mmap files


def mmap_numpy_to_boof_U8(numpy_image, boof_img = None):
    width = numpy_image.shape[1]
    height = numpy_image.shape[0]
    num_bands = 1

    mm = pyboof.mmap_file
    mm.seek(0)
    mm.write(struct.pack('>HIII', pyboof.MmapType.IMAGE_U8,width,height,num_bands))
    mm.write(numpy_image.data)

    return gateway.jvm.pyboof.PyBoofEntryPoint.mmap.readImage_U8(boof_img)


def mmap_numpy_to_boof_F32(numpy_image, boof_img = None):
    width = numpy_image.shape[1]
    height = numpy_image.shape[0]
    num_bands = 1

    mm = pyboof.mmap_file
    mm.seek(0)
    mm.write(struct.pack('>HIII', pyboof.MmapType.IMAGE_F32, width, height, num_bands))
    mm.write(numpy_image.data)

    return gateway.jvm.pyboof.PyBoofEntryPoint.mmap.readImage_F32(boof_img)


def mmap_numpy_to_boof_IU8( numpy_image , boof_img=None):
    width = numpy_image.shape[1]
    height = numpy_image.shape[0]
    num_bands = numpy_image.shape[2]

    # The image write takes less than a millisecond
    mm = pyboof.mmap_file
    mm.seek(0)
    mm.write(struct.pack('>HIII', pyboof.MmapType.IMAGE_U8, width, height, num_bands))
    mm.write(numpy_image.data)

    # TODO again just invoking the java function appears to take 2 to 3 ms even if the function does nothing
    #      Hard to tell how load the actual read takes in the MMAP file.  Probably around 1ms
    return gateway.jvm.pyboof.PyBoofEntryPoint.mmap.readImage_IU8(boof_img)


def mmap_boof_to_numpy_U8(boof_image):
    gateway.jvm.pyboof.PyBoofEntryPoint.mmap.writeImage_U8(boof_image)

    mm = pyboof.mmap_file
    mm.seek(0)
    header_bytes = mm.read(14) # speed it up significantly by minimizing disk access
    data_type, width, height, num_bands = struct.unpack('>hiii', header_bytes)

    if data_type is not pyboof.MmapType.IMAGE_U8:
        raise RuntimeError("Expected IMAGE_U8 in mmap file")
    if num_bands is not 1:
        raise RuntimeError("Expected single band image. Found {}".format(num_bands))

    data = mm.read(width*height)
    return np.ndarray(shape=(height, width), dtype=np.uint8, buffer=np.array(data))


def mmap_boof_to_numpy_F32(boof_image):
    # PERFORMANCE NOTE: Surprisingly this executes very fast.  The python code below is by far the slowest part
    gateway.jvm.pyboof.PyBoofEntryPoint.mmap.writeImage_F32(boof_image)

    mm = pyboof.mmap_file
    mm.seek(0)
    header_bytes = mm.read(14) # speed it up significantly by minimizing disk access
    data_type, width, height, num_bands = struct.unpack('>hiii', header_bytes)

    if data_type is not pyboof.MmapType.IMAGE_F32:
        raise RuntimeError("Expected IMAGE_F32 in mmap file")
    if num_bands is not 1:
        raise RuntimeError("Expected single band image. Found {}".format(num_bands))

    raw_data = mm.read(width * height * 4)
    data = struct.unpack('>{}f'.format(width*height), raw_data)

    if len(data) != width*height:
        print "Unexpected data length. {}".format(len(data))

    # create array in java format then convert into native format
    tmp = np.ndarray(shape=(height, width), dtype='>f4', order='C', buffer=raw_data)
    return tmp.astype(dtype=np.float32, copy=False)


def mmap_boof_PU8_to_numpy_IU8(boof_image):
    gateway.jvm.pyboof.PyBoofEntryPoint.mmap.writeImage_PU8_as_IU8(boof_image)

    mm = pyboof.mmap_file
    mm.seek(0)
    header_bytes = mm.read(14) # speed it up significantly by minimizing disk access
    data_type, width, height, num_bands = struct.unpack('>hiii', header_bytes)

    if data_type is not pyboof.MmapType.IMAGE_U8:
        raise RuntimeError("Expected IMAGE_U8 in mmap file")

    data = mm.read(width*height*num_bands)
    return np.ndarray(shape=(height, width, num_bands), dtype=np.uint8, buffer=np.array(data))


from image import *
from ip import *
from geo import real_nparray_to_ejml
from abc import ABCMeta, abstractmethod


class CameraModel:
    __metaclass__ = ABCMeta

    @abstractmethod
    def load(self, file_name):
        pass

    @abstractmethod
    def save(self, file_name):
        pass

    @abstractmethod
    def set_from_boof(self, boof_intrinsic):
        pass

    @abstractmethod
    def convert_to_boof(self, storage=None):
        pass


# TODO Turn this into JavaConfig instead?
class CameraPinhole(CameraModel):
    """
    BoofCV Intrinsic Camera parameters
    """
    def __init__(self):
        # Intrinsic calibration matrix
        self.fx = 0
        self.fy = 0
        self.skew = 0
        self.cx = 0
        self.cy = 0
        # image shape
        self.width = 0
        self.height = 0
        # radial distortion
        self.radial = None
        # tangential terms
        self.t1 = 0
        self.t2 = 0

    def load(self, file_name):
        file_path = os.path.abspath(file_name)
        boof_intrinsic = gateway.jvm.boofcv.io.calibration.CalibrationIO.load(file_path)

        if boof_intrinsic is None:
            raise RuntimeError("Can't load intrinsic parameters")

        self.set_from_boof(boof_intrinsic)

    def save(self, file_name):
        file_path = os.path.abspath(file_name)
        java_obj = self.convert_to_boof()
        gateway.jvm.boofcv.io.calibration.CalibrationIO.save(java_obj, file_path)

    def set_matrix(self, fx, fy, skew, cx, cy):
        self.fx = fx
        self.fy = fy
        self.skew = skew
        self.cx = cx
        self.cy = cy

    def set_image_shape(self, width, height):
        self.width = width
        self.height = height

    def set_distortion(self, radial=None, t1=0, t2=0):
        """
        Sets distortion values

        :param radial: Radial distortion
        :type radial: [float] or None
        :param t1: Tangential distortion
        :type t1: float
        :param t2:  Tangential distortion
        :type t2: float
        """
        self.radial = radial
        self.t1 = t1
        self.t2 = t2

    def set(self, orig):
        self.fx = orig.fx
        self.fy = orig.fy
        self.skew = orig.skew
        self.cx = orig.cx
        self.cy = orig.cy
        self.width = orig.width
        self.height = orig.height
        self.radial = orig.radial
        self.t1 = orig.t1
        self.t2 = orig.t2

    def set_from_boof(self, boof_intrinsic):
        self.fx = boof_intrinsic.getFx()
        self.fy = boof_intrinsic.getFy()
        self.cx = boof_intrinsic.getCx()
        self.cy = boof_intrinsic.getCy()
        self.skew = boof_intrinsic.getSkew()
        self.width = boof_intrinsic.getWidth()
        self.height = boof_intrinsic.getHeight()
        jarray = boof_intrinsic.getRadial()
        if jarray is None:
            self.radial = None
        else:
            self.radial = [float(x) for x in jarray]
        self.t1 = boof_intrinsic.getT1()
        self.t2 = boof_intrinsic.getT2()

    def convert_to_boof(self, storage=None):
        if storage is None:
            boof_intrinsic = gateway.jvm.boofcv.struct.calib.CameraPinholeRadial()
        else:
            boof_intrinsic = storage
        boof_intrinsic.setFx(float(self.fx))
        boof_intrinsic.setFy(float(self.fy))
        boof_intrinsic.setCx(float(self.cx))
        boof_intrinsic.setCy(float(self.cy))
        boof_intrinsic.setSkew(float(self.skew))
        boof_intrinsic.setWidth(int(self.width))
        boof_intrinsic.setHeight(int(self.height))
        if self.radial is not None:
            jarray = gateway.new_array(gateway.jvm.double, len(self.radial))
            for i in range(len(self.radial)):
                jarray[i] = self.radial[i]
            boof_intrinsic.setRadial(jarray)
        boof_intrinsic.setT1(float(self.t1))
        boof_intrinsic.setT2(float(self.t2))
        return boof_intrinsic

    def is_distorted(self):
        return (self.radial is not None) or self.t1 != 0 or self.t2 != 0

    def __str__(self):
        out = "Intrinsic{{ fx={:f} fy={:f} skew={:f} cx={:f} cy={:f} | width={:d} height={:d} ".\
            format(self.fx,self.fy,self.skew,self.cx,self.cy,self.width,self.height)
        if self.is_distorted():
            out += " | radial="+str(self.radial)+" t1="+str(self.t1)+" t1="+str(self.t2)+" }"
        else:
            out += "}}"
        return out


class CameraUniversalOmni(CameraPinhole):
    def __init__(self):
        CameraPinhole.__init__(self)
        self.mirror_offset = 0

    def set_from_boof(self, boof_intrinsic):
        CameraPinhole.set_from_boof(self, boof_intrinsic)
        self.mirror_offset = boof_intrinsic.getMirrorOffset()

    def convert_to_boof(self, storage=None):
        if storage is None:
            boof_intrinsic = gateway.jvm.boofcv.struct.calib.CameraUniversalOmni(0)
        else:
            boof_intrinsic = storage
        CameraPinhole.convert_to_boof(self, boof_intrinsic)
        boof_intrinsic.setMirrorOffset(float(self.mirror_offset))
        return boof_intrinsic


class LensNarrowDistortionFactory(JavaWrapper):
    def __init__(self, java_object, use_32=True):
        JavaWrapper.__init__(self, java_object)
        self.use_32 = use_32

    def distort(self, pixel_in, pixel_out):
        """

        :param pixel_in:
        :type pixel_in: bool
        :param pixel_out:
        :type pixel_in: bool
        :return: Point2Transform2_F32 or Point2Transform2_F64
        """
        if self.use_32:
            java_out = self.java_obj.distort_F32(pixel_in, pixel_out)
        else:
            java_out = self.java_obj.distort_F64(pixel_in, pixel_out)
        return Transform2to2(java_out)

    def undistort(self, pixel_in, pixel_out):
        """

        :param pixel_in:
        :type pixel_in: bool
        :param pixel_out:
        :type pixel_in: bool
        :return: Point2Transform2_F32 or Point2Transform2_F64
        """
        if self.use_32:
            java_out = self.java_obj.undistort_F32(pixel_in, pixel_out)
        else:
            java_out = self.java_obj.undistort_F32(pixel_in, pixel_out)
        return Transform2to2(java_out)

class LensWideDistortionFactory(JavaWrapper):
    def __init__(self, java_object, use_32=True):
        JavaWrapper.__init__(self, java_object)
        self.use_32 = use_32

    def distort_s_to_p(self):
        """
        Transform from unit sphere coordinates to pixel coordinates
        :return: transform
        :rtype: Transform3to2
        """
        if self.use_32:
            java_out = self.java_obj.distortStoP_F32()
        else:
            java_out = self.java_obj.distortStoP_F64()
        return Transform3to2(java_out)

    def undistort_p_to_s(self):
        """
        Transform from pixels to unit sphere coordinates
        :return: transform
        :rtype: Transform2to3
        """
        if self.use_32:
            java_out = self.java_obj.undistortPtoS_F32()
        else:
            java_out = self.java_obj.undistortPtoS_F64()
        return Transform2to3(java_out)


def create_narrow_lens_distorter( camera_model ):
    """

    :param camera_model:
    :return:
    :rtype: LensNarrowDistortionFactory
    """
    if isinstance(camera_model, CameraUniversalOmni):
        raise RuntimeError("CameraUniversalOmni is not a narrow FOV camera model")
    elif isinstance(camera_model, CameraPinhole):
        boof_model = camera_model.convert_to_boof()
        if camera_model.is_distorted():
            java_obj = gateway.jvm.boofcv.alg.distort.radtan.LensDistortionRadialTangential(boof_model)
        else:
            java_obj = gateway.jvm.boofcv.alg.distort.pinhole.LensDistortionPinhole(boof_model)
    else:
        raise RuntimeError("Unknown camera model {}".format(type(camera_model)))

    return LensNarrowDistortionFactory(java_obj)

def create_wide_lens_distorter( camera_model ):
    """

    :param camera_model:
    :return:
    :rtype: LensWideDistortionFactory
    """
    if isinstance(camera_model, CameraUniversalOmni):
        boof_model = camera_model.convert_to_boof()
        java_obj = gateway.jvm.boofcv.alg.distort.universal.LensDistortionUniversalOmni(boof_model)
    else:
        raise RuntimeError("Unknown camera model {}".format(type(camera_model)))

    return LensWideDistortionFactory(java_obj)


class NarrowToWideFovPtoP(JavaWrapper):
    """
    Distortion for converting an image from a wide FOV camera (e.g. fisheye) into a narrow FOV camera (e.g. pinhole)
    Mathematically it performs a conversion from pixels in the narrow camera to the wide camera.  The center of
    focus for the narrow camera can be changed by rotating the view by invoking set_rotation_wide_to_narrow().

    """
    def __init__(self, narrow_model, wide_model):
        """
        Constructor where camera models are specified

        :param narrow_model: Camera model for narrow FOV camera
        :type narrow_model: CameraModel
        :param wide_model: Camera model for wide FOV camera
        :type wide_model: CameraModel
        """
        narrow_distort = create_narrow_lens_distorter(narrow_model)
        wide_distort = create_wide_lens_distorter(wide_model)
        java_object = gateway.jvm.boofcv.alg.distort.NarrowToWidePtoP_F32(narrow_distort.java_obj, wide_distort.java_obj)
        JavaWrapper.__init__(self, java_object)

    def set_rotation_wide_to_narrow(self, rotation_matrix):
        """
        Used to change the principle axis of the narrow FOC camera by rotating the view

        :param rotation_matrix: 3D rotation matrix
        :return:
        """
        self.java_obj.setRotationWideToNarrow( real_nparray_to_ejml(rotation_matrix) )
        pass

    def create_image_distort(self, image_type, border_type=Border.ZERO):
        """

        :param image_type:
        :type image_type: ImageType
        :param border_type:
        :type border_type: Border
        :return: The image distort based on this transformation
        :rtype: ImageDistort
        """
        java_image_type = image_type.java_obj
        java_interp = FactoryInterpolation(image_type).bilinear(border_type=border_type)

        java_alg = gateway.jvm.boofcv.factory.distort.FactoryDistort.distort(False, java_interp, java_image_type)
        java_pixel_transform = gateway.jvm.boofcv.alg.distort.PointToPixelTransform_F32(self.java_obj)
        java_alg.setModel(java_pixel_transform)
        return ImageDistort(java_alg)


class AdjustmentType:
    NONE=0
    FULL_VIEW=1
    EXPAND=2


def adjustment_to_java(value):
    if value == AdjustmentType.NONE:
        return gateway.jvm.boofcv.alg.distort.AdjustmentType.valueOf("NONE")
    elif value == AdjustmentType.FULL_VIEW:
        return gateway.jvm.boofcv.alg.distort.AdjustmentType.valueOf("FULL_VIEW")
    elif value == AdjustmentType.EXPAND:
        return gateway.jvm.boofcv.alg.distort.AdjustmentType.valueOf("EXPAND")
    else:
        raise RuntimeError("Unknown type")


def remove_distortion(input, output, intrinsic, adjustment=AdjustmentType.FULL_VIEW, border=Border.ZERO):
    image_type = ImageType(input.getImageType())
    distorter, java_intrinsic_out = create_remove_lens_distortion(intrinsic,image_type,adjustment,border)
    distorter.apply(input,output)
    intrinsic_out = CameraPinhole()
    intrinsic_out.set_from_boof(java_intrinsic_out)
    return intrinsic_out


def create_remove_lens_distortion( intrinsic, image_type, adjustment=AdjustmentType.FULL_VIEW, border=Border.ZERO):
    java_image_type = image_type.get_java_object()
    java_adjustment = adjustment_to_java(adjustment)
    java_border = border_to_java(border)
    java_intrinsic = intrinsic.convert_to_boof()
    java_intrinsic_out = gateway.jvm.boofcv.struct.calib.CameraPinholeRadial()
    id =  gateway.jvm.boofcv.alg.distort.LensDistortionOps.imageRemoveDistortion(java_adjustment,java_border,java_intrinsic,java_intrinsic_out,java_image_type)
    return [ImageDistort(id),java_intrinsic_out]


from geo import *
from calib import *


class ConfigPolygonDetector(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.struct.Configuration.ConfigPolygonDetector")


class ConfigFiducialImage(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.factory.fiducial.ConfigFiducialImage")


class ConfigFiducialBinary(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.factory.fiducial.ConfigFiducialBinary")

class FactoryFiducial:
    def __init__(self, dtype ):
        self.boof_image_type =  dtype_to_Class_SingleBand(dtype)

    def squareImage(self, config_fiducial, config_threshold):
        """
        Creates a square image fiducial detector

        :param config_fiducial: configuration for the fiducial
        :type config_fiducial: ConfigFiducialImage
        :param config_threshold: Configuration for image thresholding step
        :type config_threshold: ThresholdType
        :return: The detector
        :rtype: FiducialImageDetector
        """
        java_detector = gateway.jvm.boofcv.factory.fiducial.FactoryFiducial.\
            squareImage(config_fiducial.java_obj, config_threshold.java_obj, self.boof_image_type)
        return FiducialImageDetector(java_detector)

    def squareBinary(self, config_fiducial, config_threshold):
        """
        Creates a binary image fiducial detector

        :param config_fiducial: configuration for the fiducial
        :type config_fiducial: ConfigFiducialBinary
        :param config_threshold: Configuration for image thresholding step
        :type config_threshold: ThresholdType
        :return: The detector
        :rtype: FiducialImageDetector
        """
        java_detector = gateway.jvm.boofcv.factory.fiducial.FactoryFiducial.\
            squareBinary(config_fiducial.java_obj, config_threshold.java_obj, self.boof_image_type)
        return FiducialDetector(java_detector)

    def chessboard(self):
        pass

    def squareGrid(self):
        pass


class FiducialDetector(JavaWrapper):
    """
    Detects fiducials and estimates their ID and 3D pose
    Wrapper around BoofCV class of the same name
    """

    def __init__(self, java_FiducialDetector):
        self.set_java_object(java_FiducialDetector)

    def detect(self, image):
        self.java_obj.detect(image)

    def set_intrinsic(self, intrinsic):
        distortion = create_narrow_lens_distorter(intrinsic)
        self.java_obj.setLensDistortion(distortion.java_obj)

    def get_total(self):
        return self.java_obj.totalFound()

    def get_fiducial_to_camera(self, which):
        """
        Returns the rigid body transform from the fiducial to the camera

        :param which: Index of detected fiducial.
        :type which: int
        :return: Rigid Body Transform
        :rtype: Se3_F64
        """
        fid_to_cam = Se3_F64()
        self.java_obj.getFiducialToCamera(which, fid_to_cam.get_java_object())
        return fid_to_cam

    def get_image_location(self, which):
        location = gateway.jvm.georegression.struct.point.Point2D_F64()
        self.java_obj.getImageLocation(which, location)
        return location.getX(), location.getY()

    def get_id(self, which):
        return self.java_obj.getId(which)

    def get_width(self, which):
        return self.java_obj.getWidth(which)

    def get_input_type(self):
        return ImageType(self.java_obj.getInputType())


class FiducialImageDetector(FiducialDetector):

    def add_pattern(self, image, side_length, threshold=100.0):
        self.java_obj.addPatternImage(image, threshold, side_length)


class ConfigCirculant(JavaConfig):
    def __init__(self, java_object=None):
        JavaConfig.__init__(self, java_object)


class ConfigTld(JavaConfig):
    def __init__(self, obj=None):
        """
        :param obj: Java object, bool, None
            If True then it will create a stable variant.  False for fast but less stable.
            None for stable
            ConfigTld for user specified configuration
        """
        if obj is None:
            config = gateway.jvm.boofcv.abst.tracker.ConfigTld()
        elif type(obj) is bool:
            config = gateway.jvm.boofcv.abst.tracker.ConfigTld(obj)
        else:
            config = obj
            JavaConfig.__init__(self, config)


class ConfigMeanShiftComaniciu(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.abst.tracker.ConfigComaniciu2003")


class FactoryTrackerObjectQuad:
    def __init__(self, image_type ):
        """
        Creates a factory for a specific image type.
        :param image_type: Specifies the type of image it processes.  Can be a dtype or ImageType
        :type image_type: int | ImageType
        """
        if isinstance(image_type, ImageType):
            self.image_type = image_type
        else:
            self.image_type = ImageType(dtype_to_ImageType(image_type))

    def circulant(self, config=None ):
        """
        Creates a Circulant tracker
        :param config: Configuration for tracker or None to use default
        :type config: None | ConfigCirculant
        :return: Tracker
        :rtype: TrackerObjectQuad
        """
        boof_image_class = self.image_type.java_obj.getImageClass()
        java_tracker = gateway.jvm.boofcv.factory.tracker.\
            FactoryTrackerObjectQuad.circulant(config, boof_image_class)
        return TrackerObjectQuad(java_tracker)

    def tld(self, config=None ):
        """
        Creates a TLD tracker
        :param config: Configuration for tracker or None to use default
        :type config: None | ConfigTld
        :return: Tracker
        :rtype: TrackerObjectQuad
        """
        boof_image_class = self.image_type.java_obj.getImageClass()
        if config is None:
            java_conf = None
        else:
            java_conf = config.java_obj
        java_tracker = gateway.jvm.boofcv.factory.tracker.\
            FactoryTrackerObjectQuad.tld(java_conf, boof_image_class)
        return TrackerObjectQuad(java_tracker)

    def mean_shift_comaniciu(self, config=None ):
        """
        Creates a Comaniciu (histogram based) style Mean-Shift tracker
        :param config: Configuration for tracker or None to use default
        :type config: None | ConfigMeanShiftComaniciu
        :return: Tracker
        :rtype: TrackerObjectQuad
        """
        if config is None:
            java_conf = None
        else:
            java_conf = config.java_obj
        java_tracker = gateway.jvm.boofcv.factory.tracker.FactoryTrackerObjectQuad.\
            meanShiftComaniciu2003(java_conf, self.image_type.java_obj)
        return TrackerObjectQuad(java_tracker)


class TrackerObjectQuad(JavaWrapper):
    """
    High level object tracker.  Takes in a quadrilateral for the initial location of the target then proceeds to
    update it for each new image in the sequence
    """
    def __init__(self, java_TrackerObjectQuad):
        JavaWrapper.__init__(self,java_TrackerObjectQuad)

    def initialize(self, image , location ):
        """
        Initialize the tracker by specifying the location of the target inside the image

        :param image: BoofCV image
        :param location: Quadrilateral2D
           Specifies the location inside the quadrilateral
        :return: bool
           True if initialization was successful or False if it failed
        """
        boof_quad = location.convert_to_boof()
        return self.java_obj.initialize( image , boof_quad )

    def process(self, image, location ):
        """
        Updates the target's location using the next image in the sequence.

        :param image: BoofCV image
        :param location: Quadrilateral2D
           (output) Will contain the new location of the tracked object
        :return: bool
           True if tracking was successful or False if it failed
        """
        boof_quad = location.convert_to_boof()
        success = self.java_obj.process( image , boof_quad )
        if success:
            location.set(boof_quad)
        return success

    def getImageType(self):
        return ImageType(self.java_obj.getImageType())


#!/usr/bin/env python

import unittest

from pyboof import gateway
import pyboof as pb
import numpy as np
import os

pb.init_memmap()


class ChecksCameraPinhole(unittest.TestCase):

    def test_setters(self):
        calib = pb.CameraPinhole()
        calib.set_image_shape(200, 300)
        calib.set_matrix(100, 101, 102, 103, 104)
        calib.set_distortion([0.1, 0.2], 103, 104)

        self.check_expected_values(calib)

    def check_expected_values(self, calib):
        self.assertEquals(calib.width, 200)
        self.assertEquals(calib.height, 300)

        self.assertEquals(calib.fx, 100)
        self.assertEquals(calib.fy, 101)
        self.assertEquals(calib.skew, 102)
        self.assertEquals(calib.cx, 103)
        self.assertEquals(calib.cy, 104)

        self.assertEquals(calib.radial, [0.1, 0.2])
        self.assertEquals(calib.t1, 103)
        self.assertEquals(calib.t2, 104)

    def test_save_load(self):
        calib = pb.CameraPinhole()
        calib.set_image_shape(200, 300)
        calib.set_matrix(100, 101, 102, 103, 104)
        calib.set_distortion([0.1, 0.2], 103, 104)

        calib.save("junk")
        calib = pb.CameraPinhole()
        calib.load("junk")
        os.remove("junk")
        self.check_expected_values(calib)


class CameraUniversalOmni(unittest.TestCase):
    def test_setters(self):
        calib = pb.CameraUniversalOmni()
        calib.set_image_shape(200, 300)
        calib.set_matrix(100, 101, 102, 103, 104)
        calib.set_distortion([0.1, 0.2], 103, 104)
        calib.mirror_offset = 3.1

        self.check_expected_values(calib)

    def check_expected_values(self, calib):
        self.assertEquals(calib.width, 200)
        self.assertEquals(calib.height, 300)

        self.assertEquals(calib.fx, 100)
        self.assertEquals(calib.fy, 101)
        self.assertEquals(calib.skew, 102)
        self.assertEquals(calib.cx, 103)
        self.assertEquals(calib.cy, 104)

        self.assertEquals(calib.radial, [0.1, 0.2])
        self.assertEquals(calib.t1, 103)
        self.assertEquals(calib.t2, 104)

        self.assertEqual(calib.mirror_offset, 3.1)

    def test_save_load(self):
        calib = pb.CameraUniversalOmni()
        calib.set_image_shape(200, 300)
        calib.set_matrix(100, 101, 102, 103, 104)
        calib.set_distortion([0.1, 0.2], 103, 104)
        calib.mirror_offset = 3.1

        calib.save("junk")
        calib = pb.CameraUniversalOmni()
        calib.load("junk")
        os.remove("junk")
        self.check_expected_values(calib)

if __name__ == '__main__':
    unittest.main()



from image import *
import ip

class GradientFlow:
    def __init__(self, derivX, derivY):
        self.derivX = derivX
        self.derivY = derivY

    def visualize(self):
        buffered_image = gateway.jvm.boofcv.gui.image.VisualizeImageData.colorizeGradient(self.derivX,self.derivY,-1)
        gateway.jvm.boofcv.gui.image.ShowImages.showWindow(buffered_image,"Gradient")


class BoofFlow:
    def __init__(self, image):
        self.image = image

    def convert_to_ndarray(self):
        return boof_to_ndarray(self.image)

    def get_boof_image(self):
        return self.image

    def get_width(self):
        return self.image.getWidth()

    def get_height(self):
        return self.image.getHeight()

    def visualize(self, name="image"):
        gateway.jvm.boofcv.gui.image.ShowImages.showWindow(self.image,name)

    def blur_gaussian(self,sigma=-1.0,radius=1):
        blurred = self.image._createNew(self.image.getWidth(),self.image.getHeight())
        ip.blur_gaussian(self.image,blurred,sigma,radius)
        return BoofFlow(blurred)

    def blur_mean(self,radius=1):
        blurred = self.image._createNew(self.image.getWidth(),self.image.getHeight())
        ip.blur_mean(self.image,blurred,radius)
        return BoofFlow(blurred)

    def blur_median(self,radius=1):
        blurred = self.image._createNew(self.image.getWidth(),self.image.getHeight())
        ip.blur_median(self.image,blurred,radius)
        return BoofFlow(blurred)

    def gradient(self,type="sobel"):
        type_grad = gradient_type(self.image)
        derivX = create_single_band(self.get_width(),self.get_height(),type_grad)
        derivY = create_single_band(self.get_width(),self.get_height(),type_grad)
        ip.gradient(self.image,derivX,derivY,type)
        return GradientFlow(derivX,derivY)


#!/usr/bin/env python

import os.path
import time
from distutils.core import setup
from setuptools.command.build_py import build_py
from setuptools import find_packages
from subprocess import call


def check_for_command( command ):
  try:
    call([command], stdout=open(os.devnull, 'w'), stderr=open(os.devnull, 'w'))
    return True
  except:
    return False


class MyBuild(build_py):
    def run(self):
        try:
            # Create build date file.  This is used to determine if the java jar and python code are
            # compatible with each other
            f = open('pyboof/build_date.txt', 'w')
            f.write(str(int(round(time.time() * 1000))))
            f.close()
            # See if javac is available for compiling the java code
            if check_for_command("javac"):
                if call(["bash", "gradlew", "allJar"]) != 0:
                    print "Gradle build failed.  "
                    exit(1)
            else:
                print "javac cannot be found. Please install it or correct your path."
                if os.path.isfile('pyboof/PyBoof-all.jar'):
                    print "     Found a precompiled jar.  Using that"
                else:
                    exit(1)
        except Exception as e:
            print "Exception message:"
            print str(e)
            print e.message
            print
            if not os.path.isfile('pyboof/PyBoof-all.jar'):
                print "Gradle build failed AND there is no PyBoof-all.jar"
                print ""
                print "Did you run build as a regular user first?"
                print "    ./setup.py build"
                print "Is Gradle and Java installed?  Test by typing the following:"
                print "    gradle allJar"
                exit(1)
        build_py.run(self)

setup(name='PyBoof',
      cmdclass={'build_py': MyBuild},
      version='0.25r1',
      description='Py4J Python wrapper for BoofCV',
      long_description=open('README.md', 'r').read(),
      url='https://github.com/lessthanoptimal/PyBoof',
      author="Peter Abeles",
      author_email="peter.abeles@gmail.com",
      eager_resources=['java'],
      packages=['pyboof'],
      package_data={'pyboof': ['PyBoof-all.jar', 'build_date.txt']},
      install_requires=['py4j==0.10.3', 'numpy>=1.11.0'],
      )



from pyboof import gateway

from image import dtype_to_Class_SingleBand
from image import ImageType
from geo import *
from py4j.java_gateway import is_instance_of


class Border:
    SKIP=0,
    EXTENDED=1
    NORMALIZED=2
    REFLECT=3
    WRAP=4
    ZERO=5


class GradientType:
    SOBEL="sobel"
    PREWITT="prewitt"
    THREE="three"
    TWO0="two0"
    TWO1="two1"


class ThresholdType:
    FIXED          = gateway.jvm.boofcv.factory.filter.binary.ThresholdType.FIXED
    GLOBAL_ENTROPY = gateway.jvm.boofcv.factory.filter.binary.ThresholdType.GLOBAL_ENTROPY
    GLOBAL_OTSU    = gateway.jvm.boofcv.factory.filter.binary.ThresholdType.GLOBAL_OTSU
    LOCAL_GAUSSIAN = gateway.jvm.boofcv.factory.filter.binary.ThresholdType.LOCAL_GAUSSIAN
    LOCAL_SQUARE   = gateway.jvm.boofcv.factory.filter.binary.ThresholdType.LOCAL_SQUARE
    LOCAL_SAVOLA   = gateway.jvm.boofcv.factory.filter.binary.ThresholdType.LOCAL_SAVOLA


class InterpolationType:
    NEAREST_NEIGHBOR=0,
    BILINEAR=1,
    BICUBIC=2,
    POLYNOMIAL4=3


class ConfigThreshold(JavaConfig):
    def __init__(self):
        JavaConfig.__init__(self,"boofcv.factory.filter.binary.ConfigThreshold")

    @staticmethod
    def create_fixed( threshold ):
        java_object = gateway.jvm.boofcv.factory.filter.binary.ConfigThreshold.fixed(float(threshold))
        return JavaConfig(java_object)

    @staticmethod
    def create_global( type ):
        java_object = gateway.jvm.pyboof.PyBoofEntryPoint.createGlobalThreshold(type)
        return JavaConfig(java_object)

    @staticmethod
    def create_local( type , radius ):
        java_object = gateway.jvm.boofcv.factory.filter.binary.ConfigThreshold.local(type,int(radius))
        return JavaConfig(java_object)


def interpolation_type_to_java( type ):
    if type == InterpolationType.NEAREST_NEIGHBOR:
        return gateway.jvm.boofcv.alg.interpolate.TypeInterpolate.NEAREST_NEIGHBOR
    elif type == InterpolationType.BICUBIC:
        return gateway.jvm.boofcv.alg.interpolate.TypeInterpolate.BICUBIC
    elif type == InterpolationType.BILINEAR:
        return gateway.jvm.boofcv.alg.interpolate.TypeInterpolate.BILINEAR
    elif type == InterpolationType.POLYNOMIAL4:
        return gateway.jvm.boofcv.alg.interpolate.TypeInterpolate.POLYNOMIAL4
    else:
        raise RuntimeError("Unknown interpolation type")


def border_to_java( border ):
    """

    :param border:
    :type border: Border
    :return: java_object
    """
    if border == Border.SKIP:
        return gateway.jvm.boofcv.core.image.border.BorderType.valueOf("SKIP")
    elif border == Border.EXTENDED:
        return gateway.jvm.boofcv.core.image.border.BorderType.valueOf("EXTENDED")
    elif border == Border.NORMALIZED:
        return gateway.jvm.boofcv.core.image.border.BorderType.valueOf("NORMALIZED")
    elif border == Border.REFLECT:
        return gateway.jvm.boofcv.core.image.border.BorderType.valueOf("REFLECT")
    elif border == Border.WRAP:
        return gateway.jvm.boofcv.core.image.border.BorderType.valueOf("WRAP")
    elif border == Border.ZERO:
        return gateway.jvm.boofcv.core.image.border.BorderType.valueOf("ZERO")


def blur_gaussian(input,output,sigma=-1.0,radius=1):
    gateway.jvm.boofcv.alg.filter.blur.BlurImageOps.gaussian(input,output,sigma,radius,None)

def blur_mean(input,output,radius=1):
    gateway.jvm.boofcv.alg.filter.blur.BlurImageOps.mean(input,output,radius,None)

def blur_median(input,output,radius=1):
    gateway.jvm.boofcv.alg.filter.blur.BlurImageOps.median(input,output,radius,None)


def gradient(input, derivX , derivY, type=GradientType.SOBEL, border=Border.EXTENDED):
    java_border = border_to_java(border)
    java_DerivativeOps = gateway.jvm.boofcv.alg.filter.derivative.GImageDerivativeOps
    java_DerivativeType = gateway.jvm.boofcv.alg.filter.derivative.DerivativeType
    if type is GradientType.SOBEL:
        java_DerivativeOps.gradient(java_DerivativeType.SOBEL,input,derivX,derivY,java_border)
    elif type is GradientType.PREWITT:
        java_DerivativeOps.gradient(java_DerivativeType.PREWITT,input,derivX,derivY,java_border)
    elif type is GradientType.THREE:
        java_DerivativeOps.gradient(java_DerivativeType.THREE,input,derivX,derivY,java_border)
    elif type is GradientType.TWO0:
        java_DerivativeOps.gradient(java_DerivativeType.TWO_0,input,derivX,derivY,java_border)
    elif type is GradientType.TWO1:
        java_DerivativeOps.gradient(java_DerivativeType.TWO_1,input,derivX,derivY,java_border)
    else:
        raise RuntimeError("Unknown gradient type "+type)


class Transform2to2(JavaWrapper):
    def __init__(self, java_object):
        JavaWrapper.__init__(self, java_object)
        if is_instance_of(gateway, java_object, gateway.jvm.boofcv.struct.distort.Point2Transform2_F32):
            self.is32 = True
            self.point_out = create_java_point_2D_f32()
        elif is_instance_of(gateway, java_object, gateway.jvm.boofcv.struct.distort.Point2Transform2_F64):
            self.is32 = False
            self.point_out = create_java_point_2D_f64()
        else:
            raise RuntimeError("Unexpected java object. "+java_object.getClass().getSimpleName())

    def apply(self, input, output=None):
        if output is None:
            output = [0.0, 0.0]
        self.java_obj.compute(float(input[0]), float(input[1]), self.point_out)
        output[0] = self.point_out.getX()
        output[1] = self.point_out.getY()
        return output


class Transform2to3(JavaWrapper):
    def __init__(self, java_object):
        JavaWrapper.__init__(self, java_object)
        if is_instance_of(gateway, java_object, gateway.jvm.boofcv.struct.distort.Point2Transform3_F32):
            self.is32 = True
            self.point_out = create_java_point_3D_f32()
        elif is_instance_of(gateway, java_object, gateway.jvm.boofcv.struct.distort.Point2Transform3_F64):
            self.is32 = False
            self.point_out = create_java_point_3D_f64()
        else:
            raise RuntimeError("Unexpected java object. "+java_object.getClass().getSimpleName())

    def apply(self, input, output=None):
        if output is None:
            output = [0.0, 0.0, 0.0]
        self.java_obj.compute(float(input[0]), float(input[1]), self.point_out)
        output[0] = self.point_out.getX()
        output[1] = self.point_out.getY()
        output[2] = self.point_out.getZ()
        return output


class Transform3to2(JavaWrapper):
    def __init__(self, java_object):
        JavaWrapper.__init__(self, java_object)
        if is_instance_of(gateway, java_object, gateway.jvm.boofcv.struct.distort.Point3Transform2_F32):
            self.is32 = True
            self.point_out = create_java_point_2D_f32()
        elif is_instance_of(gateway, java_object, gateway.jvm.boofcv.struct.distort.Point3Transform2_F64):
            self.is32 = False
            self.point_out = create_java_point_2D_f64()
        else:
            raise RuntimeError("Unexpected java object. "+java_object.getClass().getSimpleName())

    def apply(self, input, output=None):
        if output is None:
            output = [0.0, 0.0]
        self.java_obj.compute(float(input[0]), float(input[1]), float(input[2]), self.point_out)
        output[0] = self.point_out.getX()
        output[1] = self.point_out.getY()
        return output


class ImageDistort(JavaWrapper):
    """
    Applies a distortion to a BoofCV image.
    Wrapper around BoofCV ImageDistort class
    """

    def __init__(self, boof_ImageDistort):
        JavaWrapper.__init__(self, boof_ImageDistort)

    def apply(self, imageA , imageB ):
        self.java_obj.apply(imageA,imageB)


class InputToBinary(JavaWrapper):

    def __init__(self, java_object):
        self.set_java_object(java_object)

    def process(self, input , output):
        self.java_obj.process(input,output)


class FactoryThresholdBinary:
    def __init__(self, dtype ):
        self.boof_image_type =  dtype_to_Class_SingleBand(dtype)

    def localGaussian(self, radius, scale=0.95, down=True):
        """
        Create an instance of local gaussian threshold

        :param radius: Radius of local region
        :type radius: int
        :param scale: Threshold scale adjustment
        :type scale: float
        :param down: True for thresholding down and false for up
        :type down: bool
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            localGaussian(int(radius),float(scale),down,self.boof_image_type)
        return InputToBinary(java_object)

    def localSauvola(self, radius, k=0.3, down=True):
        """
        Create an instance of local gaussian threshold

        :param radius: Radius of local region
        :type radius: int
        :param k: User specified threshold adjustment factor.  Must be positive. Try 0.3
        :type k: float
        :param down: True for thresholding down and false for up
        :type down: bool
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            localSauvola(int(radius),float(k),down,self.boof_image_type)
        return InputToBinary(java_object)

    def localSquare(self, radius, scale=0.95, down=True):
        """
        Create an instance of local square threshold

        :param radius: Radius of local region
        :type radius: int
        :param scale: Threshold scale adjustment
        :type scale: float
        :param down: True for thresholding down and false for up
        :type down: bool
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            localSquare(int(radius),float(scale),down,self.boof_image_type)
        return InputToBinary(java_object)

    def globalEntropy(self, min_value=0, max_value=255, down=True):
        """
        Applies a global entropy based threshold to the entire image.

        :param min_value: Minimum pixel value.
        :type min_value: int
        :param max_value: Maximum pixel value
        :type min_value: int
        :param down: True for thresholding down and false for up
        :type down: bool:
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            globalEntropy(int(min_value),int(max_value),down,self.boof_image_type)
        return InputToBinary(java_object)

    def globalFixed(self, threshold, down=True):
        """
        Applies a fixed threshold to the entire image.

        :param min_value: Minimum pixel value.
        :type min_value: int
        :param max_value: Maximum pixel value
        :type min_value: int
        :param down: True for thresholding down and false for up
        :type down: bool:
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            globalFixed(float(threshold),down,self.boof_image_type)
        return InputToBinary(java_object)

    def globalOtsu(self, min_value=0, max_value=255, down=True):
        """
        Computes the Otsu threshold and applies it to the entire image.

        :param min_value: Minimum pixel value.
        :type min_value: int
        :param max_value: Maximum pixel value
        :type min_value: int
        :param down: True for thresholding down and false for up
        :type down: bool:
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            globalOtsu(int(min_value),int(max_value),down,self.boof_image_type)
        return InputToBinary(java_object)
    
    def threshold(self, config ):
        """
        Creates an instance of InputToBinary from a configuration class for a specific method
        :param config: Config class for a
        :type config: ConfigThreshold
        :return: New instance of InputToBinary
        :rtype: InputToBinary
        """
        java_object = gateway.jvm.boofcv.factory.filter.binary.FactoryThresholdBinary.\
            threshold(config.java_obj,self.boof_image_type)
        return InputToBinary(java_object)


class FactoryInterpolation:
    def __init__(self, image_type ):
        """

        :param image_type:
        :type image_type: ImageType
        """
        self.image_type = image_type

    def bilinear(self, min_pixel=0, max_pixel=255, border_type=Border.ZERO):
        """
        :param min_pixel:
        :type min_pixel: float
        :param max_pixel:
        :type max_pixel: float
        :param border_type:
        :type border_type: Border
        :return: java_object
        """
        java_border = border_to_java(border_type)
        java_interp = interpolation_type_to_java(InterpolationType.BILINEAR)

        return gateway.jvm.boofcv.factory.interpolate.FactoryInterpolation.\
            createPixel(float(min_pixel), float(max_pixel), java_interp, java_border, self.image_type.java_obj)



import os

import pyboof
from pyboof import ClassSingleBand_to_dtype
from pyboof import JavaWrapper
from pyboof import dtype_to_Class_SingleBand
from pyboof import gateway


class StereoParameters:
    """
    Calibration parameters for a stereo camera system.
    """
    def __init__(self, java_object = None ):
        self.left = pyboof.CameraPinhole()
        self.right = pyboof.CameraPinhole()
        self.right_to_left = pyboof.Se3_F64()

        if java_object is not None:
            self.set_from_boof(java_object)

    def load(self, path_to_file ):
        abs_path_to_file = os.path.abspath(path_to_file)
        boof_stereo = gateway.jvm.boofcv.io.calibration.CalibrationIO.load(abs_path_to_file)

        if boof_stereo is None:
            raise RuntimeError("Can't load stereo parameters")

        self.set_from_boof(boof_stereo)

    def set_from_boof(self, boof_stereo_param ):
        self.left.set_from_boof(boof_stereo_param.getLeft())
        self.right.set_from_boof(boof_stereo_param.getRight())
        self.right_to_left = pyboof.Se3_F64(boof_stereo_param.getRightToLeft())

    def convert_to_boof(self):
        boof = gateway.jvm.boofcv.struct.calib.StereoParameters()
        boof.getLeft().set(self.left.convert_to_boof())
        boof.getRight().set(self.right.convert_to_boof())
        boof.getRightToLeft().set(self.right_to_left.java_obj)

class DisparityAlgorithms:
    """
    Types of algorithms available for computing disparity
    """
    RECT = 0,
    FIVE_RECT = 1


class ConfigStereoDisparity:
    def __init__(self):
        # Which algorithm it should use
        self.type = DisparityAlgorithms.FIVE_RECT

        # Minimum disparity that it will check. Must be >= 0 and &lt; maxDisparity
        self.minDisparity = 0
        # Maximum disparity that it will calculate. Must be > 0
        self.maxDisparity = 40
        # Radius of the rectangular region along x-axis.
        self.regionRadiusX = 5
        # Radius of the rectangular region along y-axis.
        self.regionRadiusY = 5
        # Maximum allowed error in a region per pixel.  Set to <= 0 to disable.
        self.maxPerPixelError = 25
        # Tolerance for how difference the left to right associated values can be.
        self.validateRtoL = 1
        # Tolerance for how similar optimal region is to other region.  Closer to zero is more tolerant.
        self.texture = 0.2
        # Should a sub-pixel algorthm be used?
        self.subPixel = True


class StereoRectification:
    """
    Used to compute distortion for rectified stereo images
    """
    def __init__(self, intrinsic_left , intrinsic_right , right_to_left ):
        """
        Configures rectification

        :param intrinsic_left:  Intrinsic parameters for left camera
        :type intrinsic_left: pyboof.CameraPinhole
        :param intrinsic_right: Intrinsic parameters for right camera
        :type intrinsic_right: pyboof.CameraPinhole
        :param right_to_left: Extrinsic parameters for right to left camera
        :type right_to_left: pyboof.Se3_F64
        """

        boof_left = intrinsic_left.convert_to_boof()
        boof_right = intrinsic_right.convert_to_boof()

        K1 = gateway.jvm.boofcv.alg.geo.PerspectiveOps.calibrationMatrix(boof_left,None)
        K2 = gateway.jvm.boofcv.alg.geo.PerspectiveOps.calibrationMatrix(boof_right,None)
        left_to_right = right_to_left.invert()

        rectify_alg = gateway.jvm.boofcv.alg.geo.RectifyImageOps.createCalibrated()

        rectify_alg.process(K1, pyboof.Se3_F64().java_obj, K2, left_to_right.get_java_object())

        self.intrinsic_left = intrinsic_left
        self.intrinsic_right = intrinsic_right

        self.orig_rect1 = rectify_alg.getRect1()
        self.orig_rect2 = rectify_alg.getRect2()
        self.orig_rectK = rectify_alg.getCalibrationMatrix()

        self.rect1 = self.orig_rect1.copy()
        self.rect2 = self.orig_rect2.copy()
        self.rectK = self.orig_rectK.copy()

    def all_inside_left(self):
        """
        Adjusts the rectification to ensure that there are no dead regions with no pixels.
        """
        self.rect1.set(self.orig_rect1)
        self.rect2.set(self.orig_rect2)
        self.rectK.set(self.orig_rectK)

        boof_left = self.intrinsic_left.convert_to_boof()
        gateway.jvm.boofcv.alg.geo.RectifyImageOps.allInsideLeft(boof_left, self.rect1, self.rect2, self.rectK)

    def full_view_left(self):
        """
        Adjusts the rectification to ensure that the full view (every single pixel) is inside the left camera view
        """
        self.rect1.set(self.orig_rect1)
        self.rect2.set(self.orig_rect2)
        self.rectK.set(self.orig_rectK)

        boof_left = self.intrinsic_left.convert_to_boof()
        gateway.jvm.boofcv.alg.geo.RectifyImageOps.fullViewLeft(boof_left, self.rect1, self.rect2, self.rectK)

    def create_distortion(self, image_type, is_left_image):
        """
        Creates and returns a class for distorting the left image and rectifying it

        :param image_type: Type of image the distortion will process
        :type image_type: pyboof.ImageType
        :param is_left_image: If true the distortion is for the left image if false then the right image
        :type is_left_image: bool
        :return: ImageDistort class
        :rtype: pyboof.ImageDistort
        """
        boof_image_type = image_type.java_obj
        boof_border = pyboof.border_to_java( pyboof.Border.SKIP)

        if is_left_image:
            boof_distorter = gateway.jvm.boofcv.alg.geo.RectifyImageOps.\
                rectifyImage(self.intrinsic_left.convert_to_boof(), self.rect1, boof_border, boof_image_type)
        else:
            boof_distorter = gateway.jvm.boofcv.alg.geo.RectifyImageOps. \
                rectifyImage(self.intrinsic_right.convert_to_boof(), self.rect2, boof_border, boof_image_type)
        return pyboof.ImageDistort(boof_distorter)


class StereoDisparity(JavaWrapper):
    """
    Class which computes the disparity between two stereo images.  Input images are assumed to be already
    rectified.
    """
    def __init__(self, java_object ):
        JavaWrapper.__init__(self, java_object)

    def process(self, image_left, image_right):
        """
        Computes disparity from two images in BoofCV format.  To get results call
        :param image_left: BoofCV image rectified from left camera
        :param image_right: BoofCV image rectified from right camera
        """
        self.java_obj.process(image_left, image_right)

    def get_disparity_image(self):
        """
        Returns the disparity image.

        For pixel level precision a GrayU8 image is returned.  For sub-pixel a GrayF32 is returned.  Disparity
        values have a range of 0 to max-min-1 disparity.  Invalid values are any value above max-min.

        :return: BoofCV GrayU8 or GrayF32
        """
        return self.java_obj.getDisparity()

    def getBorderX(self):
        return self.java_obj.getBorderX()

    def getBorderY(self):
        return self.java_obj.getBorderY()

    def getInputType(self):
        return ClassSingleBand_to_dtype(self.java_obj.getInputType())

    def getDisparityType(self):
        return ClassSingleBand_to_dtype(self.java_obj.getDisparityType())


class FactoryStereoDisparity:
    def __init__(self, dtype ):
        self.boof_image_type =  dtype_to_Class_SingleBand(dtype)

    def region_wta(self, config):
        """
        Creates a rectangular region based winner takes all (wta) stereo disparity algorithm.
        :param config: Configuration for disparity computation
        :type config: pyboof.ConfigStereoDisparity
        :return: StereoDisparity
        :rtype: pyboof.StereoDisparity
        """
        if config is None:
            config = ConfigStereoDisparity()

        if config.type == DisparityAlgorithms.FIVE_RECT:
            alg_type = gateway.jvm.boofcv.factory.feature.disparity.DisparityAlgorithms.RECT_FIVE
        elif config.type == DisparityAlgorithms.RECT:
            alg_type = gateway.jvm.boofcv.factory.feature.disparity.DisparityAlgorithms.RECT
        else:
            raise RuntimeError("Unknown algorithm type")

        if config.subPixel:
            java_obj = gateway.jvm.boofcv.factory.feature.disparity.FactoryStereoDisparity. \
                regionSubpixelWta(alg_type, int(config.minDisparity), int(config.maxDisparity),
                                  int(config.regionRadiusX), int(config.regionRadiusY),float(config.maxPerPixelError),
                                  int(config.validateRtoL), float(config.texture), self.boof_image_type)
        else:
            java_obj = gateway.jvm.boofcv.factory.feature.disparity.FactoryStereoDisparity. \
                regionWta(alg_type, int(config.minDisparity), int(config.maxDisparity),
                          int(config.regionRadiusX), int(config.regionRadiusY), float(config.maxPerPixelError),
                          int(config.validateRtoL), float(config.texture), self.boof_image_type)

        return StereoDisparity(java_obj)




import math
import numpy as np
import py4j.java_gateway as jg
from common import *


def real_ejml_to_nparray( ejml ):
    num_rows = ejml.getNumRows()
    num_cols = ejml.getNumCols()

    M = np.zeros((num_rows,num_cols))
    for i in xrange(num_rows):
        for j in xrange(num_cols):
            M[i,j] = ejml.unsafe_get(i,j)
    return M

def real_nparray_to_ejml( array ):
    num_rows = array.shape[0]
    num_cols = array.shape[1]

    M = gateway.jvm.org.ejml.data.DenseMatrix64F(num_rows,num_cols)
    for i in xrange(num_rows):
        for j in xrange(num_cols):
            M.unsafe_set(i,j,array[i,j])
    return M

class Se3_F64(JavaWrapper):
    def __init__(self, java_Se3F64=None):
        if java_Se3F64 is None:
            JavaWrapper.__init__(self, gateway.jvm.georegression.struct.se.Se3_F64())
        else:
            JavaWrapper.__init__(self, java_Se3F64)

    def invert(self):
        return Se3_F64(self.java_obj.invert(None))

    def get_rotation(self):
        return real_ejml_to_nparray(self.java_obj.getR())

    def get_translation(self):
        T = self.java_obj.getT()
        return (T.getX(),T.getY(),T.getZ())


def create_java_point_2D_f32(x=0., y=0.):
    return gateway.jvm.georegression.struct.point.Point2D_F32(float(x), float(y))


def create_java_point_2D_f64(x=0., y=0.):
    return gateway.jvm.georegression.struct.point.Point2D_F64(float(x), float(y))


def create_java_point_3D_f32(x=0., y=0., z=0.):
    return gateway.jvm.georegression.struct.point.Point3D_F32(float(x), float(y), float(z))


def create_java_point_3D_f64(x=0., y=0., z=0.):
    return gateway.jvm.georegression.struct.point.Point3D_F64(float(x), float(y), float(z))


def tuple_to_Point2D_F64(ptuple, jpoint=None):
    if jpoint == None:
        return create_java_point_2D_f32(ptuple[0],ptuple[1])
    else:
        jpoint.set(float(ptuple[0]),float(ptuple[1]))
        return jpoint


def tuple_to_Point2D_F32(ptuple ,jpoint=None):
    if jpoint == None:
        return create_java_point_2D_f64(ptuple[0],ptuple[1])
    else:
        jpoint.set(float(ptuple[0]),float(ptuple[1]))
        return jpoint


class Point2D:
    def __init__(self, x=0, y=0):
        """
        :param x: float
            x-coordinate
        :param y: float
            y-coordinate
        """
        self.x = x
        self.y = y

    def convert_to_boof(self):
        return create_java_point_2D_f64(float(self.x), float(self.y))

    def set(self, o):
        if type(o) is Point2D:
            self.x = o.x
            self.y = o.y
        if type(o) is tuple:
            self.x = o[0]
            self.y = o[1]
        elif jg.is_instance_of(gateway, o, gateway.jvm.georegression.struct.point.Point2D_F64):
            self.x = o.getX()
            self.y = o.getY()
        else:
            raise Exception("Unknown object type")

    def distance(self, point):
        dx = point.x-self.x
        dy = point.y-self.y
        return math.sqrt(dx*dx + dy*dy)


    def get_tuple(self):
        """
        Returns the values of the point inside a tuple: (x,y)
        """
        return (self.x,self.y)

    def get_x(self):
        return self.x

    def get_y(self):
        return self.y

    def set_x(self, x):
        self.x = x

    def set_y(self,y ):
        self.y = y


class Quadrilateral2D:
    """
    Four sided polygon specified by its vertexes.  The vertexes are ordered, but it's not specified if they are
    ordered in clockwise or counter clockwise direction
    """
    def __init__(self, a=Point2D(), b=Point2D(), c=Point2D(), d=Point2D()):
        """

        :param a: Point2D
            Vertex is the quadrilateral
        :param b: Point2D
            Vertex is the quadrilateral
        :param c: Point2D
            Vertex is the quadrilateral
        :param d: Point2D
            Vertex is the quadrilateral
        """
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    def convert_to_boof(self):
        a = self.a.convert_to_boof()
        b = self.b.convert_to_boof()
        c = self.c.convert_to_boof()
        d = self.d.convert_to_boof()

        return gateway.jvm.georegression.struct.shapes.Quadrilateral_F64(a,b,c,d)

    def set(self, o):
        if type(o) is Quadrilateral2D:
            self.a.set(o.a)
            self.b.set(o.b)
            self.c.set(o.c)
            self.d.set(o.d)

        elif jg.is_instance_of(gateway, o, gateway.jvm.georegression.struct.shapes.Quadrilateral_F64 ):
            self.a.set(o.getA())
            self.b.set(o.getB())
            self.c.set(o.getC())
            self.d.set(o.getD())
        else:
            raise Exception("Unknown object type")

    def get_vertexes(self):
        """
        Returns a tuple with all the vertexes (a,b,c,d)
        :return: tuple of all vertexes
        """
        return (self.a,self.b,self.c,self.d)

    def get_tuple_tuple(self):
        return (self.a.get_tuple(),self.b.get_tuple(),self.c.get_tuple(),self.d.get_tuple())

    def get_a(self):
        return self.a

    def get_b(self):
        return self.b

    def get_c(self):
        return self.c

    def get_d(self):
        return self.d
